from __future__ import annotations as _annotations

import asyncio
import contextlib
import json
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from http.client import HTTPConnection, HTTPSConnection
from typing import Any, cast
from urllib.parse import urlsplit, urlunsplit

from acp import connect_to_agent
from acp.client.connection import ClientSideConnection
from acp.interfaces import Client
from websockets.asyncio.client import ClientConnection, connect

from .auth import bearer_headers
from .config import TransportOptions
from .limits import DEFAULT_OPEN_TIMEOUT
from .metadata import ServerMetadata
from .stream import WebSocketStreamBridge, open_websocket_stream_bridge

__all__ = ("RemoteClientConnection", "connect_remote_agent", "fetch_server_metadata")

_HeaderPairs = Mapping[str, str] | Sequence[tuple[str, str]]


@dataclass(slots=True)
class RemoteClientConnection:
    connection: ClientSideConnection
    websocket: ClientConnection
    streams: WebSocketStreamBridge
    metadata: ServerMetadata | None = None

    async def close(self) -> None:
        try:
            await self.connection.close()
        finally:
            await self.streams.close()


async def connect_remote_agent(
    client: Client,
    url: str,
    *,
    options: TransportOptions | None = None,
    headers: _HeaderPairs | None = None,
    bearer_token: str | None = None,
) -> RemoteClientConnection:
    resolved_options = options or TransportOptions()
    resolved_headers = _merge_headers(headers, bearer_headers(bearer_token))
    websocket = await connect(
        url,
        additional_headers=resolved_headers,
        compression=resolved_options.compression,
        open_timeout=resolved_options.open_timeout,
        ping_interval=resolved_options.ping_interval,
        ping_timeout=resolved_options.ping_timeout,
        close_timeout=resolved_options.close_timeout,
        max_size=resolved_options.max_size,
        max_queue=resolved_options.max_queue,
    )
    streams: WebSocketStreamBridge | None = None
    connection: ClientSideConnection | None = None
    try:
        streams = await open_websocket_stream_bridge(
            websocket,
            reader_limit=resolved_options.reader_limit,
        )
        connection = connect_to_agent(client, streams.writer, streams.reader)
        metadata = await fetch_server_metadata(
            url,
            headers=resolved_headers,
            timeout=resolved_options.open_timeout,
        )
    except BaseException:
        if connection is not None:
            with contextlib.suppress(BaseException):
                await connection.close()
        if streams is not None:
            with contextlib.suppress(BaseException):
                await streams.close()
        else:
            with contextlib.suppress(BaseException):
                await websocket.close()
            with contextlib.suppress(BaseException):
                await websocket.wait_closed()
        raise
    return RemoteClientConnection(
        connection=connection,
        websocket=websocket,
        streams=streams,
        metadata=metadata,
    )


async def fetch_server_metadata(
    url: str,
    *,
    headers: _HeaderPairs | None = None,
    timeout: float = DEFAULT_OPEN_TIMEOUT,
) -> ServerMetadata | None:
    metadata_url = _metadata_url(url)
    if metadata_url is None:
        return None
    return await asyncio.to_thread(_fetch_server_metadata_sync, metadata_url, headers, timeout)


def _merge_headers(
    base: _HeaderPairs | None,
    extra: dict[str, str] | None,
) -> _HeaderPairs | None:
    if not extra:
        return base
    if base is None:
        return extra
    if isinstance(base, Mapping):
        base_mapping = cast("Mapping[str, str]", base)
        merged_items: list[tuple[str, str]] = [*base_mapping.items(), *extra.items()]
        return merged_items
    base_pairs = cast("Sequence[tuple[str, str]]", base)
    merged_pairs: list[tuple[str, str]] = [*base_pairs, *extra.items()]
    return merged_pairs


def _metadata_url(url: str) -> str | None:
    parsed = urlsplit(url)
    if parsed.scheme not in {"ws", "wss"}:
        return None
    path = parsed.path or "/"
    if not path.endswith("/ws"):
        return None
    metadata_path = path.removesuffix("/ws") or "/"
    scheme = "https" if parsed.scheme == "wss" else "http"
    return urlunsplit((scheme, parsed.netloc, metadata_path, "", ""))


def _fetch_server_metadata_sync(
    url: str,
    headers: _HeaderPairs | None,
    timeout: float = DEFAULT_OPEN_TIMEOUT,
) -> ServerMetadata | None:
    parsed = urlsplit(url)
    if parsed.hostname is None:
        return None
    connection_class = HTTPSConnection if parsed.scheme == "https" else HTTPConnection
    connection = connection_class(parsed.hostname, parsed.port, timeout=timeout)
    try:
        connection.request("GET", parsed.path or "/", headers=_headers_dict(headers))
        response = connection.getresponse()
        if response.status != 200:
            return None
        payload = json.loads(response.read().decode("utf-8"))
    except (OSError, ValueError, json.JSONDecodeError):
        return None
    finally:
        connection.close()
    if not isinstance(payload, dict):
        return None
    try:
        return ServerMetadata(
            transport_kind=str(payload["transport_kind"]),
            transport_version=int(payload["transport_version"]),
            package_version=str(payload["package_version"]),
            auth_required=bool(payload["auth_required"]),
            supported_auth_modes=tuple(_string_list(payload.get("supported_auth_modes"))),
            max_size=int(payload["max_size"]),
            max_queue=int(payload["max_queue"]),
            compression=_optional_str(payload.get("compression")),
            health_path=str(payload["health_path"]),
            metadata_path=str(payload["metadata_path"]),
            websocket_path=str(payload["websocket_path"]),
            supported_agent_families=tuple(_string_list(payload.get("supported_agent_families"))),
            remote_cwd=_optional_str(payload.get("remote_cwd")),
        )
    except (KeyError, TypeError, ValueError):
        return None


def _headers_dict(headers: _HeaderPairs | None) -> dict[str, str]:
    if headers is None:
        return {}
    if isinstance(headers, Mapping):
        return dict(headers)
    return dict(headers)


def _string_list(value: Any) -> list[str]:
    if not isinstance(value, list):
        return []
    return [item for item in value if isinstance(item, str)]


def _optional_str(value: Any) -> str | None:
    if value is None:
        return None
    if isinstance(value, str):
        return value
    return str(value)



from __future__ import annotations as _annotations

from .base import BufferedCapabilityBridge, CapabilityBridge
from .capability_support import (
    AnthropicCompactionBridge,
    HarnessCodeModeBridge,
    HarnessFileSystemBridge,
    HarnessShellBridge,
    ImageGenerationBridge,
    IncludeToolReturnSchemasBridge,
    McpCapabilityBridge,
    OpenAICompactionBridge,
    PrefixToolsBridge,
    SetToolMetadataBridge,
    ThreadExecutorBridge,
    ToolsetBridge,
    WebFetchBridge,
    WebSearchBridge,
)
from .external_hooks import EventEmissionMode, ExternalHookEventBridge
from .history_processor import (
    HistoryProcessorBridge,
    HistoryProcessorCallable,
    HistoryProcessorContextual,
    HistoryProcessorPlain,
    HistoryProcessorWithContextAsync,
    HistoryProcessorWithContextSync,
)
from .hooks import HookBridge
from .mcp import McpBridge, McpServerDefinition, McpToolDefinition, SessionMcpBridge
from .prepare_tools import (
    PlanGenerationType,
    PrepareOutputToolsBridge,
    PrepareOutputToolsMode,
    PrepareToolsBridge,
    PrepareToolsMode,
)
from .thinking import ThinkingBridge

__all__ = (
    "AnthropicCompactionBridge",
    "BufferedCapabilityBridge",
    "CapabilityBridge",
    "EventEmissionMode",
    "ExternalHookEventBridge",
    "HarnessCodeModeBridge",
    "HarnessFileSystemBridge",
    "HarnessShellBridge",
    "HistoryProcessorBridge",
    "HistoryProcessorCallable",
    "HistoryProcessorContextual",
    "HistoryProcessorPlain",
    "HistoryProcessorWithContextAsync",
    "HistoryProcessorWithContextSync",
    "HookBridge",
    "ImageGenerationBridge",
    "IncludeToolReturnSchemasBridge",
    "McpBridge",
    "McpCapabilityBridge",
    "McpServerDefinition",
    "McpToolDefinition",
    "OpenAICompactionBridge",
    "PlanGenerationType",
    "PrefixToolsBridge",
    "PrepareOutputToolsBridge",
    "PrepareOutputToolsMode",
    "PrepareToolsBridge",
    "PrepareToolsMode",
    "SetToolMetadataBridge",
    "SessionMcpBridge",
    "ThinkingBridge",
    "ThreadExecutorBridge",
    "ToolsetBridge",
    "WebFetchBridge",
    "WebSearchBridge",
)



from __future__ import annotations as _annotations

import base64
from collections.abc import Sequence
from dataclasses import dataclass, replace
from typing import Any, Literal, TypeAlias, cast

from acp.schema import (
    AudioContentBlock,
    BlobResourceContents,
    EmbeddedResourceContentBlock,
    ImageContentBlock,
    ResourceContentBlock,
    TextContentBlock,
    TextResourceContents,
    Usage,
    UserMessageChunk,
)
from pydantic_ai import (
    AgentRunResult,
    ModelMessagesTypeAdapter,
    ModelRequest,
    ModelResponse,
    RunUsage,
    TextPart,
)
from pydantic_ai.messages import (
    AudioUrl,
    BinaryContent,
    BinaryImage,
    DocumentUrl,
    ImageUrl,
    ModelMessage,
    RetryPromptPart,
    ToolCallPart,
    ToolReturnPart,
    UserContent,
    UserPromptPart,
)
from pydantic_ai.tools import DeferredToolRequests

PromptBlock: TypeAlias = (
    AudioContentBlock
    | EmbeddedResourceContentBlock
    | ImageContentBlock
    | ResourceContentBlock
    | TextContentBlock
)
PromptInput: TypeAlias = str | list[UserContent]

__all__ = (
    "PromptBlock",
    "PromptInput",
    "PromptRunOutcome",
    "build_cancelled_history",
    "build_user_updates",
    "contains_deferred_tool_requests",
    "derive_title",
    "dump_message_history",
    "load_message_history",
    "prompt_to_input",
    "prompt_to_text",
    "sanitize_message_history",
    "usage_from_run",
)


@dataclass(slots=True, kw_only=True)
class PromptRunOutcome:
    result: AgentRunResult[Any]
    stop_reason: Literal["cancelled", "end_turn"]
    streamed_output: bool = False


def build_user_updates(
    prompt: Sequence[PromptBlock],
    *,
    message_id: str,
) -> list[UserMessageChunk]:
    return [
        UserMessageChunk(
            session_update="user_message_chunk",
            content=block,
            message_id=message_id,
        )
        for block in prompt
    ]


def derive_title(prompt: Sequence[PromptBlock]) -> str:
    text = prompt_to_text(prompt).strip()
    if not text:
        return "Untitled session"
    collapsed = " ".join(text.split())
    return collapsed[:80]


def load_message_history(raw_history: str | None) -> list[ModelMessage]:
    if raw_history is None:
        return []
    return sanitize_message_history(ModelMessagesTypeAdapter.validate_json(raw_history))


def dump_message_history(messages: list[ModelMessage]) -> str:
    return ModelMessagesTypeAdapter.dump_json(messages).decode("utf-8")


def contains_deferred_tool_requests(output_type: Any) -> bool:
    if output_type is DeferredToolRequests:
        return True
    if isinstance(output_type, Sequence):
        return any(contains_deferred_tool_requests(item) for item in output_type)
    return False


def sanitize_message_history(
    messages: list[ModelMessage],
    *,
    error_text: str | None = None,
) -> list[ModelMessage]:
    unresolved_tool_calls = _find_unprocessed_tool_calls(messages)
    if not unresolved_tool_calls:
        return list(messages)

    unresolved_ids = {part.tool_call_id for part in unresolved_tool_calls}
    sanitized_messages: list[ModelMessage] = []
    for message in messages:
        if not isinstance(message, ModelResponse):
            sanitized_messages.append(message)
            continue
        filtered_parts = [
            part
            for part in message.parts
            if not isinstance(part, ToolCallPart) or part.tool_call_id not in unresolved_ids
        ]
        if filtered_parts:
            sanitized_messages.append(replace(message, parts=filtered_parts))

    sanitized_messages.append(
        ModelResponse(
            parts=[
                TextPart(
                    _render_unprocessed_tool_call_text(
                        unresolved_tool_calls,
                        error_text=error_text,
                    ),
                ),
            ],
        ),
    )
    return sanitized_messages


def prompt_to_text(prompt: Sequence[PromptBlock]) -> str:
    parts: list[str] = []
    for block in prompt:
        if isinstance(block, TextContentBlock):
            parts.append(block.text)
        elif isinstance(block, ResourceContentBlock):
            parts.append(_format_resource_link(block.name, block.uri))
        elif isinstance(block, EmbeddedResourceContentBlock):
            resource = block.resource
            if isinstance(resource, TextResourceContents):
                parts.append(_format_text_resource_context(resource.uri, resource.text))
            else:
                parts.append(f"[embedded-resource:{resource.uri}]")
        elif isinstance(block, ImageContentBlock):
            parts.append("[image]")
        else:
            parts.append("[audio]")
    return "\n\n".join(parts)


def prompt_to_input(prompt: Sequence[PromptBlock]) -> PromptInput:
    multimodal_parts: list[UserContent] = []
    has_non_text_blocks = False
    for block in prompt:
        if isinstance(block, TextContentBlock):
            multimodal_parts.append(block.text)
            continue
        has_non_text_blocks = True
        multimodal_parts.append(_prompt_block_to_user_content(block))

    if not has_non_text_blocks:
        return prompt_to_text(prompt)
    return multimodal_parts


def build_error_history(
    raw_history: str | None,
    *,
    prompt_text: str,
    traceback_text: str,
) -> str:
    messages = load_message_history(raw_history)
    stripped_prompt = prompt_text.strip()
    if stripped_prompt:
        messages.append(ModelRequest(parts=[UserPromptPart(stripped_prompt)]))
    messages = sanitize_message_history(messages, error_text=traceback_text)
    if not _history_contains_text(messages, traceback_text):
        messages.append(
            ModelResponse(
                parts=[
                    TextPart(
                        "\n".join(
                            (
                                "The previous run failed before completion.",
                                "",
                                "Traceback:",
                                traceback_text.rstrip(),
                            ),
                        ),
                    ),
                ],
            ),
        )
    return dump_message_history(messages)


def build_cancelled_history(
    raw_history: str | None,
    *,
    prompt_text: str,
    details_text: str,
) -> str:
    messages = load_message_history(raw_history)
    stripped_prompt = prompt_text.strip()
    if stripped_prompt:
        messages.append(ModelRequest(parts=[UserPromptPart(stripped_prompt)]))
    messages = sanitize_message_history(messages)
    cancelled_text = _render_cancelled_run_text(details_text)
    if not _history_contains_text(messages, cancelled_text):
        messages.append(ModelResponse(parts=[TextPart(cancelled_text)]))
    return dump_message_history(messages)


def usage_from_run(usage: RunUsage) -> Usage | None:
    if not usage.has_values():
        return None
    thought_tokens = usage.details.get("reasoning_tokens") or usage.details.get("thought_tokens")
    return Usage(
        input_tokens=usage.input_tokens,
        output_tokens=usage.output_tokens,
        cached_read_tokens=usage.cache_read_tokens or None,
        cached_write_tokens=usage.cache_write_tokens or None,
        thought_tokens=thought_tokens,
        total_tokens=(
            usage.input_tokens
            + usage.output_tokens
            + usage.cache_read_tokens
            + usage.cache_write_tokens
            + usage.input_audio_tokens
            + usage.cache_audio_read_tokens
            + usage.output_audio_tokens
        ),
    )


def _find_unprocessed_tool_calls(messages: list[ModelMessage]) -> list[ToolCallPart]:
    seen_parts: list[ToolCallPart] = []
    resolved_tool_call_ids: set[str] = set()
    for message in messages:
        if isinstance(message, ModelResponse):
            for part in message.parts:
                if isinstance(part, ToolCallPart):
                    seen_parts.append(part)
        elif isinstance(message, ModelRequest):
            for part in message.parts:
                if isinstance(part, ToolReturnPart | RetryPromptPart):
                    resolved_tool_call_ids.add(part.tool_call_id)
    return [part for part in seen_parts if part.tool_call_id not in resolved_tool_call_ids]


def _prompt_block_to_user_content(
    block: (
        AudioContentBlock | EmbeddedResourceContentBlock | ImageContentBlock | ResourceContentBlock
    ),
) -> UserContent:
    if isinstance(block, ImageContentBlock):
        return BinaryImage(
            data=base64.b64decode(block.data),
            media_type=block.mime_type,
        )
    if isinstance(block, AudioContentBlock):
        return BinaryContent(
            data=base64.b64decode(block.data),
            media_type=block.mime_type,
        )
    if isinstance(block, ResourceContentBlock):
        return _resource_link_to_user_content(block)
    return _embedded_resource_to_user_content(block)


def _resource_link_to_user_content(block: ResourceContentBlock) -> UserContent:
    mime_type = block.mime_type
    if mime_type is None or mime_type.startswith("text/"):
        return _format_resource_link(block.name, block.uri)
    if mime_type.startswith("image/"):
        return ImageUrl(url=block.uri, media_type=mime_type)
    if mime_type.startswith("audio/"):
        return AudioUrl(url=block.uri, media_type=mime_type)
    return DocumentUrl(url=block.uri, media_type=mime_type)


def _embedded_resource_to_user_content(
    block: EmbeddedResourceContentBlock,
) -> UserContent:
    resource = block.resource
    if isinstance(resource, TextResourceContents):
        return _format_text_resource_context(resource.uri, resource.text)
    blob_resource = cast("BlobResourceContents", resource)
    return BinaryContent.narrow_type(
        BinaryContent(
            data=base64.b64decode(blob_resource.blob),
            media_type=blob_resource.mime_type or "application/octet-stream",
        ),
    )


def _format_resource_link(name: str | None, uri: str) -> str:
    if name:
        return f"[@{name}]({uri})"
    file_name = _resource_name_from_uri(uri)
    if file_name is None:
        return uri
    return f"[@{file_name}]({uri})"


def _format_text_resource_context(uri: str, text: str) -> str:
    return "\n".join(
        (
            _format_resource_link(None, uri),
            f'<context ref="{uri}">',
            text,
            "</context>",
        ),
    )


def _resource_name_from_uri(uri: str) -> str | None:
    if uri.startswith("file://"):
        path = uri.removeprefix("file://")
        name = path.split("/")[-1]
        return name or path
    if uri.startswith("zed://"):
        name = uri.rsplit("/", maxsplit=1)[-1]
        return name or uri
    return None


def _render_unprocessed_tool_call_text(
    tool_calls: list[ToolCallPart],
    *,
    error_text: str | None,
) -> str:
    lines = [
        "One or more tool calls were removed from history because they were not processed.",
        "",
        "Removed tool calls:",
    ]
    for part in tool_calls:
        lines.append(f"- {part.tool_name} ({part.tool_call_id})")
    if error_text:
        lines.extend(("", "Traceback:", error_text.rstrip()))
    return "\n".join(lines)


def _render_cancelled_run_text(details_text: str) -> str:
    return "\n".join(
        (
            "User stopped the run.",
            "",
            "Run details:",
            details_text.rstrip(),
        ),
    )


def _history_contains_text(messages: list[ModelMessage], text: str) -> bool:
    needle = text.strip()
    if not needle:
        return False
    for message in messages:
        if not isinstance(message, ModelResponse):
            continue
        for part in message.parts:
            if isinstance(part, TextPart) and needle in part.content:
                return True
    return False



from __future__ import annotations as _annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import TypeAlias

from acp.exceptions import RequestError
from acp.schema import (
    AvailableCommand,
    PlanEntry,
    SessionConfigOptionBoolean,
    SessionConfigOptionSelect,
    SessionConfigSelectOption,
    SessionModelState,
    SessionModeState,
)

from ..models import AdapterModel
from ..providers import ModelSelectionState, ModeState

ConfigOption: TypeAlias = SessionConfigOptionSelect | SessionConfigOptionBoolean

__all__ = (
    "ConfigOption",
    "SessionSurface",
    "build_mode_config_option",
    "build_mode_state_from_selection",
    "build_model_config_option",
    "build_model_state_from_selection",
    "find_model_option",
)


@dataclass(slots=True, kw_only=True)
class SessionSurface:
    config_options: list[ConfigOption] | None
    model_state: SessionModelState | None
    mode_state: SessionModeState | None
    plan_entries: list[PlanEntry] | None
    available_commands: list[AvailableCommand] | None = None


def build_model_config_option(
    model_selection_state: ModelSelectionState,
) -> SessionConfigOptionSelect:
    current_model_id = model_selection_state.current_model_id
    if current_model_id is None:
        raise RequestError.internal_error({"reason": "missing_current_model_id"})
    return SessionConfigOptionSelect(
        id="model",
        name=model_selection_state.config_option_name,
        category="model",
        description=model_selection_state.config_option_description,
        type="select",
        current_value=current_model_id,
        options=[
            SessionConfigSelectOption(
                value=model.model_id,
                name=model.name,
                description=model.description,
            )
            for model in model_selection_state.available_models
        ],
    )


def build_mode_config_option(mode_state: ModeState) -> SessionConfigOptionSelect:
    current_mode_id = mode_state.current_mode_id
    if current_mode_id is None:
        raise RequestError.internal_error({"reason": "missing_current_mode_id"})
    return SessionConfigOptionSelect(
        id="mode",
        name="Mode",
        category="mode",
        description="Session-local mode selection.",
        type="select",
        current_value=current_mode_id,
        options=[
            SessionConfigSelectOption(
                value=mode.id,
                name=mode.name,
                description=mode.description,
            )
            for mode in mode_state.modes
        ],
    )


def build_model_state_from_selection(
    model_selection_state: ModelSelectionState | None,
) -> SessionModelState | None:
    if model_selection_state is None or model_selection_state.current_model_id is None:
        return None
    return SessionModelState(
        available_models=[
            model.to_model_info() for model in model_selection_state.available_models
        ],
        current_model_id=model_selection_state.current_model_id,
    )


def build_mode_state_from_selection(
    mode_state: ModeState | None,
) -> SessionModeState | None:
    if mode_state is None or mode_state.current_mode_id is None or not mode_state.modes:
        return None
    return SessionModeState(
        available_modes=list(mode_state.modes),
        current_mode_id=mode_state.current_mode_id,
    )


def find_model_option(
    model_id: str,
    *,
    available_models: Sequence[AdapterModel],
) -> AdapterModel | None:
    for model in available_models:
        if model.model_id == model_id:
            return model
    return None



from __future__ import annotations as _annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from acp.exceptions import RequestError
from acp.schema import AgentPlanUpdate, PlanEntry
from pydantic_ai import Agent as PydanticAgent
from pydantic_ai.tools import ToolDefinition

from ..awaitables import resolve_value
from ..bridges import PrepareToolsBridge
from ..session.state import AcpSessionContext, utc_now
from ._agent_state import (
    has_native_plan_tools,
    set_native_plan_tools_installed,
    try_active_session,
)

if TYPE_CHECKING:
    from .adapter import PydanticAcpAgent

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
OutputDataT = TypeVar("OutputDataT", covariant=True)

_GET_PLAN_TOOL_NAME = "acp_get_plan"
_SET_PLAN_TOOL_NAME = "acp_set_plan"
_UPDATE_PLAN_ENTRY_TOOL_NAME = "acp_update_plan_entry"
_MARK_PLAN_DONE_TOOL_NAME = "acp_mark_plan_done"

__all__ = ("_NativePlanRuntime",)


class _NativePlanRuntime(Generic[AgentDepsT, OutputDataT]):
    def __init__(self, owner: PydanticAcpAgent[AgentDepsT, OutputDataT]) -> None:
        self._owner = owner
        self._native_plan_updates: set[str] = set()

    def native_plan_bridge(
        self,
        session: AcpSessionContext,
    ) -> PrepareToolsBridge[Any] | None:
        for bridge in self._owner._config.capability_bridges:
            if isinstance(bridge, PrepareToolsBridge) and bridge.supports_plan_tools(session):
                return bridge
        return None

    def supports_native_plan_state(self, session: AcpSessionContext) -> bool:
        return (
            self._owner._config.plan_provider is None
            and self.native_plan_bridge(session) is not None
        )

    def requires_native_plan_output(self, session: AcpSessionContext) -> bool:
        bridge = self.native_plan_bridge(session)
        if bridge is None:
            return False
        return bridge.is_plan_mode(session) and bridge.uses_structured_plan_generation(session)

    def supports_native_plan_progress(self, session: AcpSessionContext) -> bool:
        bridge = self.native_plan_bridge(session)
        if bridge is None:
            return False
        return bridge.supports_plan_progress(session)

    def supports_native_plan_writes(self, session: AcpSessionContext) -> bool:
        bridge = self.native_plan_bridge(session)
        if bridge is None:
            return False
        return bridge.supports_plan_write_tools(session)

    def get_native_plan_entries(self, session: AcpSessionContext) -> list[PlanEntry] | None:
        if not session.plan_entries:
            return None
        return [PlanEntry.model_validate(entry) for entry in session.plan_entries]

    def set_native_plan_state(
        self,
        session: AcpSessionContext,
        *,
        entries: Sequence[PlanEntry],
        plan_markdown: str | None,
    ) -> None:
        session.plan_entries = [
            entry.model_dump(mode="json", exclude_none=True) for entry in entries
        ]
        session.plan_markdown = plan_markdown

    async def persist_external_plan_state(
        self,
        session: AcpSessionContext,
        *,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        entries: Sequence[PlanEntry],
        plan_markdown: str | None,
    ) -> None:
        persistence_provider = self._owner._config.native_plan_persistence_provider
        if persistence_provider is None:
            return
        await resolve_value(
            persistence_provider.persist_plan_state(
                session,
                agent,
                entries,
                plan_markdown,
            ),
        )

    async def emit_native_plan_update(self, session: AcpSessionContext) -> None:
        client = self._owner._client
        if client is None:
            return
        entries = self.get_native_plan_entries(session)
        if entries is None:
            return
        self._native_plan_updates.add(session.session_id)
        await client.session_update(
            session_id=session.session_id,
            update=AgentPlanUpdate(session_update="plan", entries=entries),
        )

    def consume_native_plan_update(self, session: AcpSessionContext) -> bool:
        if session.session_id not in self._native_plan_updates:
            return False
        self._native_plan_updates.remove(session.session_id)
        return True

    async def persist_native_plan_state(
        self,
        session: AcpSessionContext,
        *,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        entries: Sequence[PlanEntry],
        plan_markdown: str | None,
    ) -> None:
        self.set_native_plan_state(
            session,
            entries=entries,
            plan_markdown=plan_markdown,
        )
        await self.persist_external_plan_state(
            session,
            agent=agent,
            entries=entries,
            plan_markdown=plan_markdown,
        )
        session.updated_at = utc_now()
        self._owner._config.session_store.save(session)
        await self.emit_native_plan_update(session)

    async def persist_current_native_plan_state(
        self,
        session: AcpSessionContext,
        *,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        entries = self.get_native_plan_entries(session)
        if entries is None and session.plan_markdown is None:
            return
        await self.persist_native_plan_state(
            session,
            agent=agent,
            entries=() if entries is None else entries,
            plan_markdown=session.plan_markdown,
        )

    def replace_native_plan_entry(
        self,
        session: AcpSessionContext,
        *,
        index: int,
        status: str | None = None,
        content: str | None = None,
        priority: str | None = None,
    ) -> PlanEntry:
        entries = self.get_native_plan_entries(session)
        if not entries:
            raise RequestError.invalid_params({"plan": "No plan entries have been recorded yet."})
        if index < 1 or index > len(entries):
            raise RequestError.invalid_params(
                {
                    "index": index,
                    "plan": f"Plan entry index must be between 1 and {len(entries)}.",
                },
            )
        existing_entry = entries[index - 1]
        updated_payload = existing_entry.model_dump(mode="python")
        if status is not None:
            updated_payload["status"] = status
        if content is not None:
            updated_payload["content"] = content
        if priority is not None:
            updated_payload["priority"] = priority
        updated_entry = PlanEntry.model_validate(updated_payload)
        entries[index - 1] = updated_entry
        self.set_native_plan_state(
            session,
            entries=entries,
            plan_markdown=session.plan_markdown,
        )
        return updated_entry

    def format_native_plan(self, session: AcpSessionContext) -> str:
        entries = self.get_native_plan_entries(session)
        additional_instructions = self._native_plan_additional_instructions()
        if not entries:
            if session.plan_markdown:
                if additional_instructions is None:
                    return session.plan_markdown
                return "\n\n".join(
                    (
                        session.plan_markdown.rstrip(),
                        "Additional plan instructions:",
                        additional_instructions,
                    ),
                )
            return "No plan has been recorded yet."
        numbered_entries = "\n".join(
            f"{index}. [{entry.status}] ({entry.priority}) {entry.content}"
            for index, entry in enumerate(entries, start=1)
        )
        index_guidance = (
            "Use these 1-based entry numbers with "
            f"`{_UPDATE_PLAN_ENTRY_TOOL_NAME}` and `{_MARK_PLAN_DONE_TOOL_NAME}`."
        )
        sections = [index_guidance]
        if additional_instructions is not None:
            sections.extend(("Additional plan instructions:", additional_instructions))
        if not session.plan_markdown:
            sections.append(numbered_entries)
            return "\n\n".join(sections)
        return "\n\n".join(
            (
                session.plan_markdown.rstrip(),
                "Current plan entries:",
                numbered_entries,
                *sections,
            ),
        )

    def _native_plan_additional_instructions(self) -> str | None:
        instructions = self._owner._config.native_plan_additional_instructions
        if instructions is None:
            return None
        normalized = instructions.strip()
        if normalized == "":
            return None
        return normalized

    def install_native_plan_tools(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        if self._owner._config.plan_provider is not None:
            return
        if has_native_plan_tools(agent):
            return

        def prepare_plan_access_tool(ctx: Any, tool_def: ToolDefinition) -> ToolDefinition | None:
            del ctx
            active_session = try_active_session(agent)
            if active_session is None:
                return None
            if not self.supports_native_plan_state(active_session):
                return None
            return tool_def

        def prepare_plan_progress_tool(ctx: Any, tool_def: ToolDefinition) -> ToolDefinition | None:
            prepared_tool = prepare_plan_access_tool(ctx, tool_def)
            if prepared_tool is None:
                return None
            active_session = try_active_session(agent)
            if active_session is None:
                return None
            if not self.supports_native_plan_progress(active_session):
                return None
            return prepared_tool

        def prepare_plan_write_tool(ctx: Any, tool_def: ToolDefinition) -> ToolDefinition | None:
            prepared_tool = prepare_plan_access_tool(ctx, tool_def)
            if prepared_tool is None:
                return None
            active_session = try_active_session(agent)
            if active_session is None:
                return None
            if not self.supports_native_plan_writes(active_session):
                return None
            return prepared_tool

        tool_plain = agent.tool_plain

        @tool_plain(name=_GET_PLAN_TOOL_NAME, prepare=prepare_plan_access_tool)
        def acp_get_plan() -> str:
            """Return the saved plan and numbered entries.

            The returned entry numbers are 1-based. Use those same numbers with
            `acp_update_plan_entry` and `acp_mark_plan_done`.
            """
            active_session = try_active_session(agent)
            if active_session is None:
                return "No active ACP session is bound."
            return self.format_native_plan(active_session)

        @tool_plain(name=_SET_PLAN_TOOL_NAME, prepare=prepare_plan_write_tool)
        async def acp_set_plan(entries: list[PlanEntry], plan_md: str | None = None) -> str:
            """Replace the current plan state with the provided entries."""
            active_session = try_active_session(agent)
            if active_session is None:
                return "No active ACP session is bound."
            await self.persist_native_plan_state(
                active_session,
                agent=agent,
                entries=entries,
                plan_markdown=plan_md,
            )
            return f"Recorded {len(entries)} plan entries."

        @tool_plain(name=_UPDATE_PLAN_ENTRY_TOOL_NAME, prepare=prepare_plan_progress_tool)
        async def acp_update_plan_entry(
            index: int,
            status: str | None = None,
            content: str | None = None,
            priority: str | None = None,
        ) -> str:
            """Update a single plan entry by its 1-based index."""
            active_session = try_active_session(agent)
            if active_session is None:
                return "No active ACP session is bound."
            updated_entry = self.replace_native_plan_entry(
                active_session,
                index=index,
                status=status,
                content=content,
                priority=priority,
            )
            await self.persist_current_native_plan_state(active_session, agent=agent)
            return (
                f"Updated plan entry {index}: "
                f"[{updated_entry.status}] ({updated_entry.priority}) {updated_entry.content}"
            )

        @tool_plain(name=_MARK_PLAN_DONE_TOOL_NAME, prepare=prepare_plan_progress_tool)
        async def acp_mark_plan_done(index: int) -> str:
            """Mark a single plan entry completed by its 1-based index."""
            active_session = try_active_session(agent)
            if active_session is None:
                return "No active ACP session is bound."
            updated_entry = self.replace_native_plan_entry(
                active_session,
                index=index,
                status="completed",
            )
            await self.persist_current_native_plan_state(active_session, agent=agent)
            return f"Marked plan entry {index} as completed: {updated_entry.content}"

        set_native_plan_tools_installed(agent)



from __future__ import annotations as _annotations

from collections.abc import Awaitable
from dataclasses import dataclass
from typing import Generic, Protocol, TypeVar

from pydantic_ai import Agent as PydanticAgent

from .awaitables import resolve_value
from .session.state import AcpSessionContext

AgentFactoryDepsT = TypeVar("AgentFactoryDepsT", contravariant=True)
AgentFactoryOutputDataT = TypeVar("AgentFactoryOutputDataT", covariant=True)
AgentDepsT = TypeVar("AgentDepsT")
OutputDataT = TypeVar("OutputDataT")

__all__ = ("AgentFactory", "AgentSource", "FactoryAgentSource", "StaticAgentSource")


class AgentFactory(Protocol[AgentFactoryDepsT, AgentFactoryOutputDataT]):
    def __call__(
        self,
        session: AcpSessionContext,
    ) -> (
        PydanticAgent[AgentFactoryDepsT, AgentFactoryOutputDataT]
        | Awaitable[PydanticAgent[AgentFactoryDepsT, AgentFactoryOutputDataT]]
    ): ...


class AgentSource(Protocol[AgentDepsT, OutputDataT]):
    async def get_agent(
        self,
        session: AcpSessionContext,
    ) -> PydanticAgent[AgentDepsT, OutputDataT]: ...

    async def get_deps(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> AgentDepsT | None: ...


@dataclass(slots=True)
class StaticAgentSource(Generic[AgentDepsT, OutputDataT]):
    agent: PydanticAgent[AgentDepsT, OutputDataT]
    deps: AgentDepsT | None = None

    async def get_agent(self, session: AcpSessionContext) -> PydanticAgent[AgentDepsT, OutputDataT]:
        del session
        return self.agent

    async def get_deps(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> AgentDepsT | None:
        del session, agent
        return self.deps


@dataclass(slots=True)
class FactoryAgentSource(Generic[AgentDepsT, OutputDataT]):
    factory: AgentFactory[AgentDepsT, OutputDataT]

    async def get_agent(self, session: AcpSessionContext) -> PydanticAgent[AgentDepsT, OutputDataT]:
        return await resolve_value(self.factory(session))

    async def get_deps(
        self,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> AgentDepsT | None:
        del session, agent
        return None



from __future__ import annotations as _annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from acp.schema import (
    McpCapabilities,
    ToolCallProgress,
    ToolCallStart,
    ToolCallStatus,
    ToolKind,
)
from pydantic_ai.capabilities import AbstractCapability
from pydantic_ai.settings import ModelSettings

from ..agent_types import RuntimeAgent
from ..providers import ConfigOption, ModeState
from ..session.state import AcpSessionContext, JsonValue, SessionTranscriptUpdate

if TYPE_CHECKING:
    from ..projection import ProjectionMap

__all__ = (
    "BufferedCapabilityBridge",
    "CapabilityBridge",
)


class CapabilityBridge:
    metadata_key: str | None = None

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...] | None:
        del session
        return None

    def drain_updates(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> list[SessionTranscriptUpdate] | None:
        del session, agent
        return None

    def get_config_options(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> list[ConfigOption] | None:
        del session, agent
        return None

    def get_mcp_capabilities(self, agent: RuntimeAgent | None = None) -> McpCapabilities | None:
        del agent
        return None

    def get_projection_maps(self) -> tuple[ProjectionMap, ...]:
        return ()

    def get_approval_policy_key(
        self,
        tool_name: str,
        raw_input: JsonValue | None = None,
    ) -> str | None:
        del tool_name, raw_input
        return None

    def get_mode_state(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> ModeState | None:
        del session, agent
        return None

    def get_model_settings(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> ModelSettings | None:
        del session, agent
        return None

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue] | None:
        del session, agent
        return None

    def get_tool_kind(self, tool_name: str, raw_input: JsonValue | None = None) -> ToolKind | None:
        del tool_name, raw_input
        return None

    def set_config_option(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        config_id: str,
        value: str | bool,
    ) -> list[ConfigOption] | None:
        del session, agent, config_id, value
        return None

    def set_mode(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        mode_id: str,
    ) -> ModeState | None:
        del session, agent, mode_id
        return None


@dataclass(slots=True)
class BufferedCapabilityBridge(CapabilityBridge):
    _event_counts: dict[str, int] = field(default_factory=dict, init=False, repr=False)
    _pending_updates: dict[str, list[SessionTranscriptUpdate]] = field(
        default_factory=dict,
        init=False,
        repr=False,
    )

    def drain_updates(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> list[SessionTranscriptUpdate] | None:
        del agent
        pending = self._pending_updates.pop(session.session_id, None)
        return list(pending) if pending else None

    def _record_completed_event(
        self,
        session: AcpSessionContext,
        *,
        title: str,
        kind: ToolKind = "execute",
        raw_input: JsonValue | None = None,
        raw_output: str | None = None,
    ) -> None:
        event_id = self._record_started_event(
            session,
            title=title,
            kind=kind,
            raw_input=raw_input,
        )
        self._record_progress_event(
            session,
            event_id=event_id,
            title=title,
            kind=kind,
            status="completed",
            raw_output=raw_output,
        )

    def _record_failed_event(
        self,
        session: AcpSessionContext,
        *,
        title: str,
        kind: ToolKind = "execute",
        raw_input: JsonValue | None = None,
        raw_output: str | None = None,
    ) -> None:
        event_id = self._record_started_event(
            session,
            title=title,
            kind=kind,
            raw_input=raw_input,
        )
        self._record_progress_event(
            session,
            event_id=event_id,
            title=title,
            kind=kind,
            status="failed",
            raw_output=raw_output,
        )

    def _record_started_event(
        self,
        session: AcpSessionContext,
        *,
        title: str,
        kind: ToolKind = "execute",
        raw_input: JsonValue | None = None,
    ) -> str:
        event_id = self._next_event_id(session)
        self._append_updates(
            session,
            [
                ToolCallStart(
                    session_update="tool_call",
                    tool_call_id=event_id,
                    title=title,
                    kind=kind,
                    status="in_progress",
                    raw_input=raw_input,
                ),
            ],
        )
        return event_id

    def _record_progress_event(
        self,
        session: AcpSessionContext,
        *,
        event_id: str,
        title: str,
        status: ToolCallStatus,
        kind: ToolKind = "execute",
        raw_output: str | None = None,
    ) -> None:
        self._append_updates(
            session,
            [
                ToolCallProgress(
                    session_update="tool_call_update",
                    tool_call_id=event_id,
                    title=title,
                    kind=kind,
                    status=status,
                    raw_output=raw_output,
                ),
            ],
        )

    def _append_updates(
        self,
        session: AcpSessionContext,
        updates: Sequence[SessionTranscriptUpdate],
    ) -> None:
        self._pending_updates.setdefault(session.session_id, []).extend(updates)

    def _next_event_id(self, session: AcpSessionContext) -> str:
        next_count = self._event_counts.get(session.session_id, 0) + 1
        self._event_counts[session.session_id] = next_count
        bridge_name = type(self).__name__.replace("Bridge", "").lower() or "bridge"
        return f"{session.session_id}:{bridge_name}:{next_count}"



from __future__ import annotations as _annotations

from dataclasses import dataclass, field
from inspect import Parameter, signature
from typing import Protocol

from acp.exceptions import RequestError
from acp.interfaces import Client as AcpClient
from acp.schema import PermissionOption
from pydantic_ai.messages import ToolCallPart
from pydantic_ai.tools import (
    DeferredToolRequests,
    DeferredToolResults,
    ToolApproved,
    ToolDenied,
)
from typing_extensions import TypeIs

from .approval_store import (
    ApprovalPolicy,
    ApprovalPolicyStore,
    PermissionOptionSet,
    SessionMetadataApprovalPolicyStore,
)
from .permission_presentation import (
    DefaultPermissionToolCallBuilder,
    PermissionRequestContext,
    PermissionToolCallBuilder,
)
from .projection import ProjectionMap, ToolClassifier
from .session.state import AcpSessionContext, JsonValue

__all__ = (
    "ApprovalBridge",
    "ApprovalPolicy",
    "ApprovalPolicyStore",
    "ApprovalResolution",
    "NativeApprovalBridge",
    "PermissionOptionSet",
    "ProjectionAwareApprovalBridge",
    "SessionMetadataApprovalPolicyStore",
    "supports_projection_aware_approval_bridge",
)


@dataclass(slots=True, frozen=True, kw_only=True)
class ApprovalResolution:
    deferred_tool_results: DeferredToolResults
    cancelled: bool = False
    cancelled_tool_call: ToolCallPart | None = None


class ApprovalBridge(Protocol):
    async def resolve_deferred_approvals(
        self,
        *,
        client: AcpClient,
        session: AcpSessionContext,
        requests: DeferredToolRequests,
        classifier: ToolClassifier,
    ) -> ApprovalResolution: ...


class ProjectionAwareApprovalBridge(Protocol):
    async def resolve_deferred_approvals(
        self,
        *,
        client: AcpClient,
        session: AcpSessionContext,
        requests: DeferredToolRequests,
        classifier: ToolClassifier,
        projection_map: ProjectionMap | None = None,
    ) -> ApprovalResolution: ...


def supports_projection_aware_approval_bridge(
    value: object,
) -> TypeIs[ProjectionAwareApprovalBridge]:
    resolver = getattr(value, "resolve_deferred_approvals", None)
    if not callable(resolver):
        return False
    try:
        resolver_signature = signature(resolver)
    except (TypeError, ValueError):
        return False
    parameters = resolver_signature.parameters
    if "projection_map" in parameters:
        return True
    return any(parameter.kind is Parameter.VAR_KEYWORD for parameter in parameters.values())


@dataclass(slots=True, kw_only=True)
class NativeApprovalBridge:
    enable_persistent_choices: bool = False
    tool_call_builder: PermissionToolCallBuilder = field(
        default_factory=DefaultPermissionToolCallBuilder,
    )
    policy_store: ApprovalPolicyStore = field(default_factory=SessionMetadataApprovalPolicyStore)
    option_set: PermissionOptionSet = field(default_factory=PermissionOptionSet)

    async def resolve_deferred_approvals(
        self,
        *,
        client: AcpClient,
        session: AcpSessionContext,
        requests: DeferredToolRequests,
        classifier: ToolClassifier,
        projection_map: ProjectionMap | None = None,
    ) -> ApprovalResolution:
        deferred_results = DeferredToolResults(metadata=dict(requests.metadata))
        for tool_call in requests.approvals:
            raw_input = tool_call.args_as_dict()
            approval_policy_key = classifier.approval_policy_key(tool_call.tool_name, raw_input)
            remembered_policy = self._get_remembered_policy(session, approval_policy_key)
            if remembered_policy is not None:
                deferred_results.approvals[tool_call.tool_call_id] = self._policy_to_result(
                    remembered_policy,
                )
                continue

            permission_response = await client.request_permission(
                options=self._build_permission_options(),
                session_id=session.session_id,
                tool_call=self.tool_call_builder.build_tool_call_update(
                    PermissionRequestContext(
                        session=session,
                        tool_call=tool_call,
                        raw_input=dict(raw_input),
                        cwd=session.cwd,
                        classifier=classifier,
                        projection_map=projection_map,
                    ),
                ),
            )
            outcome = permission_response.outcome
            if outcome.outcome == "cancelled":
                return ApprovalResolution(
                    deferred_tool_results=deferred_results,
                    cancelled=True,
                    cancelled_tool_call=tool_call,
                )

            selected_result = self._selected_option_to_result(outcome.option_id)
            if selected_result is None:
                raise RequestError.invalid_request({"optionId": outcome.option_id})
            self._remember_policy(
                session=session,
                approval_policy_key=approval_policy_key,
                option_id=outcome.option_id,
            )
            deferred_results.approvals[tool_call.tool_call_id] = selected_result

        return ApprovalResolution(deferred_tool_results=deferred_results)

    def _build_permission_options(self) -> list[PermissionOption]:
        options = [
            PermissionOption(
                option_id="allow_once",
                name=self.option_set.allow_once_name,
                kind="allow_once",
            ),
            PermissionOption(
                option_id="reject_once",
                name=self.option_set.reject_once_name,
                kind="reject_once",
            ),
        ]
        if not self.enable_persistent_choices:
            return options
        return [
            options[0],
            PermissionOption(
                option_id="allow_always",
                name=self.option_set.allow_always_name,
                kind="allow_always",
            ),
            options[1],
            PermissionOption(
                option_id="reject_always",
                name=self.option_set.reject_always_name,
                kind="reject_always",
            ),
        ]

    def _selected_option_to_result(
        self,
        option_id: str,
    ) -> ToolApproved | ToolDenied | None:
        if option_id in {"allow_once", "allow_always"}:
            return ToolApproved()
        if option_id in {"reject_once", "reject_always"}:
            return ToolDenied()
        return None

    def _remember_policy(
        self,
        *,
        session: AcpSessionContext,
        approval_policy_key: str,
        option_id: str,
    ) -> None:
        if not self.enable_persistent_choices:
            return
        if option_id == "allow_always":
            self._set_remembered_policy(
                session,
                approval_policy_key=approval_policy_key,
                policy="allow",
            )
        elif option_id == "reject_always":
            self._set_remembered_policy(
                session,
                approval_policy_key=approval_policy_key,
                policy="reject",
            )

    def _get_remembered_policy(
        self,
        session: AcpSessionContext,
        approval_policy_key: str,
    ) -> ApprovalPolicy | None:
        if not self.enable_persistent_choices:
            return None
        return self.policy_store.get_policy(session, approval_policy_key)

    def _set_remembered_policy(
        self,
        session: AcpSessionContext,
        *,
        approval_policy_key: str,
        policy: ApprovalPolicy,
    ) -> None:
        self.policy_store.set_policy(session, approval_policy_key, policy)

    def _approval_policies(self, session: AcpSessionContext) -> dict[str, JsonValue]:
        exported = self.policy_store.export_state(session)
        return exported if exported is not None else {}

    def _policy_to_result(self, policy: ApprovalPolicy) -> ToolApproved | ToolDenied:
        if policy == "allow":
            return ToolApproved()
        return ToolDenied()



from __future__ import annotations as _annotations

from collections.abc import Awaitable, Mapping
from dataclasses import dataclass, field
from typing import cast

from acp.schema import (
    ModelInfo,
    PlanEntry,
    PlanEntryPriority,
    PlanEntryStatus,
    SessionMode,
    ToolKind,
)

from ..providers import (
    ConfigOption,
    ConfigOptionsProvider,
    ModelSelectionState,
    ModeState,
    SessionModelsProvider,
    SessionModesProvider,
)
from ..session.state import AcpSessionContext, JsonValue
from .base import CapabilityBridge

__all__ = (
    "ConfigOptionsBridge",
    "DeepAgentsCompatibilityBridge",
    "ModeSelectionBridge",
    "ModelSelectionBridge",
    "ToolSurfaceBridge",
)


@dataclass(slots=True, frozen=True, kw_only=True)
class ToolSurfaceBridge(CapabilityBridge):
    tool_kinds: Mapping[str, ToolKind] = field(default_factory=dict)
    approval_policy_keys: Mapping[str, str] = field(default_factory=dict)

    def get_tool_kind(self, tool_name: str, raw_input: JsonValue | None = None) -> ToolKind | None:
        del raw_input
        return self.tool_kinds.get(tool_name)

    def get_approval_policy_key(
        self,
        tool_name: str,
        raw_input: JsonValue | None = None,
    ) -> str | None:
        del raw_input
        return self.approval_policy_keys.get(tool_name)


@dataclass(slots=True, frozen=True, kw_only=True)
class ModelSelectionBridge(CapabilityBridge):
    available_models: tuple[ModelInfo, ...] = ()
    default_model_id: str | None = None
    provider: SessionModelsProvider | None = None

    def get_model_state(
        self,
        session: AcpSessionContext,
    ) -> ModelSelectionState | None | Awaitable[ModelSelectionState | None]:
        if self.provider is not None:
            return self.provider.get_model_state(session=session)
        if not self.available_models:
            return None
        current_model_id = session.session_model_id or self._default_model_id()
        return ModelSelectionState(
            available_models=list(self.available_models),
            current_model_id=current_model_id,
        )

    def set_model(
        self,
        session: AcpSessionContext,
        model_id: str,
    ) -> ModelSelectionState | None | Awaitable[ModelSelectionState | None]:
        if self.provider is not None:
            return self.provider.set_model(session=session, model_id=model_id)
        if not any(model.model_id == model_id for model in self.available_models):
            return None
        session.session_model_id = model_id
        return self.get_model_state(session)

    def _default_model_id(self) -> str | None:
        if self.default_model_id is not None:
            return self.default_model_id
        if self.available_models:
            return self.available_models[0].model_id
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class ModeSelectionBridge(CapabilityBridge):
    available_modes: tuple[SessionMode, ...] = ()
    default_mode_id: str | None = None
    provider: SessionModesProvider | None = None

    def get_mode_state(
        self,
        session: AcpSessionContext,
    ) -> ModeState | None | Awaitable[ModeState | None]:
        if self.provider is not None:
            return self.provider.get_mode_state(session=session)
        if not self.available_modes:
            return None
        current_mode_id = session.session_mode_id or self._default_mode_id()
        return ModeState(
            modes=list(self.available_modes),
            current_mode_id=current_mode_id,
        )

    def set_mode(
        self,
        session: AcpSessionContext,
        mode_id: str,
    ) -> ModeState | None | Awaitable[ModeState | None]:
        if self.provider is not None:
            return self.provider.set_mode(session=session, mode_id=mode_id)
        if not any(mode.id == mode_id for mode in self.available_modes):
            return None
        session.session_mode_id = mode_id
        return self.get_mode_state(session)

    def _default_mode_id(self) -> str | None:
        if self.default_mode_id is not None:
            return self.default_mode_id
        if self.available_modes:
            return self.available_modes[0].id
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class ConfigOptionsBridge(CapabilityBridge):
    provider: ConfigOptionsProvider

    def get_config_options(
        self,
        session: AcpSessionContext,
    ) -> list[ConfigOption] | None | Awaitable[list[ConfigOption] | None]:
        return self.provider.get_config_options(session=session)

    def set_config_option(
        self,
        session: AcpSessionContext,
        config_id: str,
        value: str | bool,
    ) -> list[ConfigOption] | None | Awaitable[list[ConfigOption] | None]:
        return self.provider.set_config_option(session=session, config_id=config_id, value=value)


@dataclass(slots=True, frozen=True, kw_only=True)
class DeepAgentsCompatibilityBridge(CapabilityBridge):
    metadata_key: str | None = "deepagents"

    def get_session_metadata(self, session: AcpSessionContext) -> dict[str, JsonValue] | None:
        plan_generation_type = session.config_values.get("plan_generation_type")
        metadata: dict[str, JsonValue] = {
            "cwd": str(session.cwd),
            "mode": session.session_mode_id,
            "model": session.session_model_id,
        }
        if isinstance(plan_generation_type, str):
            metadata["plan_generation_type"] = plan_generation_type
        return metadata

    def extract_plan_entries(self, payload: JsonValue) -> list[PlanEntry] | None:
        if not isinstance(payload, dict):
            return None
        todos = payload.get("todos")
        if not isinstance(todos, list):
            return None
        entries: list[PlanEntry] = []
        for todo in todos:
            if not isinstance(todo, dict):
                continue
            content = todo.get("content")
            if not isinstance(content, str):
                continue
            status = todo.get("status", "pending")
            if status not in {"pending", "in_progress", "completed"}:
                status = "pending"
            priority = todo.get("priority", "medium")
            if priority not in {"high", "medium", "low"}:
                priority = "medium"
            entries.append(
                PlanEntry(
                    content=content,
                    status=cast("PlanEntryStatus", status),
                    priority=cast("PlanEntryPriority", priority),
                ),
            )
        return entries



from __future__ import annotations as _annotations

from dataclasses import dataclass, field
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Literal, TypeAlias, assert_never

from acp.schema import (
    AgentMessageChunk,
    AgentPlanUpdate,
    SessionInfoUpdate,
    ToolCallProgress,
    ToolCallStart,
    UserMessageChunk,
)
from pydantic import BaseModel
from typing_extensions import TypeIs

if TYPE_CHECKING:
    from acp.interfaces import Client as AcpClient

JsonPrimitive: TypeAlias = None | bool | int | float | str
JsonValue: TypeAlias = JsonPrimitive | list["JsonValue"] | dict[str, "JsonValue"]
SessionTranscriptUpdate: TypeAlias = (
    AgentMessageChunk
    | AgentPlanUpdate
    | SessionInfoUpdate
    | ToolCallProgress
    | ToolCallStart
    | UserMessageChunk
)
SessionTranscriptKind: TypeAlias = Literal[
    "agent_message_chunk",
    "plan",
    "session_info_update",
    "tool_call",
    "tool_call_update",
    "user_message_chunk",
]

_SESSION_UPDATE_MODELS: dict[str, type[BaseModel]] = {
    "agent_message_chunk": AgentMessageChunk,
    "plan": AgentPlanUpdate,
    "session_info_update": SessionInfoUpdate,
    "tool_call": ToolCallStart,
    "tool_call_update": ToolCallProgress,
    "user_message_chunk": UserMessageChunk,
}

_TRANSCRIPT_KINDS: Final = frozenset(_SESSION_UPDATE_MODELS)


def _is_transcript_kind(value: JsonValue) -> TypeIs[SessionTranscriptKind]:
    return isinstance(value, str) and value in _TRANSCRIPT_KINDS


def utc_now() -> datetime:
    return datetime.now(UTC)


@dataclass(slots=True, kw_only=True)
class StoredSessionUpdate:
    kind: SessionTranscriptKind
    payload: dict[str, JsonValue]

    @classmethod
    def from_update(cls, update: SessionTranscriptUpdate) -> StoredSessionUpdate:
        payload = _coerce_json_object(
            update.model_dump(mode="json", by_alias=True, exclude_none=True),
        )
        session_update = payload.get("sessionUpdate")
        if not _is_transcript_kind(session_update):
            raise TypeError("Session transcript update payload is missing `sessionUpdate`.")
        return cls(kind=session_update, payload=payload)

    def to_update(self) -> SessionTranscriptUpdate:
        if self.kind == "agent_message_chunk":
            return AgentMessageChunk.model_validate(self.payload)
        if self.kind == "plan":
            return AgentPlanUpdate.model_validate(self.payload)
        if self.kind == "session_info_update":
            return SessionInfoUpdate.model_validate(self.payload)
        if self.kind == "tool_call":
            return ToolCallStart.model_validate(self.payload)
        if self.kind == "tool_call_update":
            return ToolCallProgress.model_validate(self.payload)
        if self.kind == "user_message_chunk":
            return UserMessageChunk.model_validate(self.payload)
        assert_never(self.kind)


@dataclass(slots=True, kw_only=True)
class AcpSessionContext:
    session_id: str
    cwd: Path
    created_at: datetime
    updated_at: datetime
    title: str | None = None
    session_model_id: str | None = None
    session_mode_id: str | None = None
    plan_entries: list[dict[str, JsonValue]] = field(default_factory=list)
    plan_markdown: str | None = None
    config_values: dict[str, str | bool] = field(default_factory=dict)
    mcp_servers: list[dict[str, JsonValue]] = field(default_factory=list)
    metadata: dict[str, JsonValue] = field(default_factory=dict)
    transcript: list[StoredSessionUpdate] = field(default_factory=list)
    client: AcpClient | None = field(default=None, repr=False, compare=False)


def _coerce_json_object(value: Any) -> dict[str, JsonValue]:
    if not isinstance(value, dict):
        raise TypeError("Expected a JSON object payload.")
    payload: dict[str, JsonValue] = {}
    for key, item in value.items():
        if not isinstance(key, str):
            raise TypeError("JSON object keys must be strings.")
        payload[key] = _coerce_json_value(item)
    return payload


def _coerce_json_value(value: Any) -> JsonValue:
    if value is None or isinstance(value, bool | int | float | str):
        return value
    if isinstance(value, list):
        return [_coerce_json_value(item) for item in value]
    if isinstance(value, dict):
        return _coerce_json_object(value)
    raise TypeError(f"Unsupported JSON value: {type(value).__name__}")


__all__ = (
    "AcpSessionContext",
    "JsonValue",
    "SessionTranscriptUpdate",
    "StoredSessionUpdate",
    "utc_now",
)



from __future__ import annotations as _annotations

from dataclasses import dataclass
from pathlib import Path

from acp.interfaces import Client as AcpClient

from ..session.state import AcpSessionContext
from .filesystem import ClientFilesystemBackend
from .policy import HostAccessPolicy
from .terminal import ClientTerminalBackend

__all__ = ("ClientHostContext",)


@dataclass(slots=True, frozen=True, kw_only=True)
class ClientHostContext:
    client: AcpClient
    session: AcpSessionContext
    filesystem: ClientFilesystemBackend
    terminal: ClientTerminalBackend
    access_policy: HostAccessPolicy | None = None
    workspace_root: Path | None = None

    @classmethod
    def from_session(
        cls,
        *,
        client: AcpClient,
        session: AcpSessionContext,
        access_policy: HostAccessPolicy | None = None,
        workspace_root: Path | None = None,
    ) -> ClientHostContext:
        return cls(
            client=client,
            session=session,
            filesystem=ClientFilesystemBackend(
                client=client,
                session=session,
                access_policy=access_policy,
                workspace_root=workspace_root,
            ),
            terminal=ClientTerminalBackend(
                client=client,
                session=session,
                access_policy=access_policy,
                workspace_root=workspace_root,
            ),
            access_policy=access_policy,
            workspace_root=workspace_root,
        )

    @classmethod
    def from_bound_session(
        cls,
        session: AcpSessionContext,
        *,
        access_policy: HostAccessPolicy | None = None,
        workspace_root: Path | None = None,
    ) -> ClientHostContext:
        client = session.client
        if client is None:
            raise ValueError("The ACP client is not connected to this session context.")
        return cls.from_session(
            client=client,
            session=session,
            access_policy=access_policy,
            workspace_root=workspace_root,
        )



from __future__ import annotations as _annotations

import asyncio
from collections.abc import Awaitable
from typing import Any, cast

import pytest
from pydantic_ai.capabilities import PrepareOutputTools

from .support import (
    UTC,
    AcpSessionContext,
    Agent,
    Path,
    PrepareOutputToolsBridge,
    PrepareOutputToolsMode,
    PrepareToolsBridge,
    PrepareToolsMode,
    RunContext,
    SessionConfigOptionSelect,
    TestModel,
    ToolDefinition,
    datetime,
)


def _passthrough_tools(
    ctx: RunContext[None],
    tool_defs: list[ToolDefinition],
) -> list[ToolDefinition]:
    del ctx
    return list(tool_defs)


def test_passthrough_tools_helper_returns_a_copy() -> None:
    tool_defs: list[ToolDefinition] = []
    copied = _passthrough_tools(cast("Any", None), tool_defs)
    assert copied == []
    assert copied is not tool_defs


def test_prepare_output_tools_bridge_builds_capability_and_metadata(
    tmp_path: Path,
) -> None:
    def keep_public(
        ctx: RunContext[None],
        tool_defs: list[ToolDefinition],
    ) -> list[ToolDefinition]:
        del ctx
        return [tool_def for tool_def in tool_defs if tool_def.name == "public"]

    bridge = PrepareOutputToolsBridge(
        default_mode_id="default",
        modes=[
            PrepareOutputToolsMode(
                id="default",
                name="Default",
                description="Default output tools.",
                prepare_func=keep_public,
            ),
            PrepareOutputToolsMode(
                id="strict",
                name="Strict",
                description=None,
                prepare_func=_passthrough_tools,
            ),
        ],
    )
    session = AcpSessionContext(
        session_id="prepare-output-tools",
        cwd=tmp_path,
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    tool_defs = [ToolDefinition(name="public"), ToolDefinition(name="private")]

    async def run_prepare() -> list[ToolDefinition]:
        prepared = bridge.build_prepare_output_tools(session)
        result = cast("Awaitable[list[ToolDefinition]]", prepared(cast("Any", None), tool_defs))
        return await result

    capability = bridge.build_capability(session)
    contributions = bridge.build_agent_capabilities(session)
    prepared_tools = asyncio.run(run_prepare())
    metadata = bridge.get_session_metadata(session, Agent(TestModel()))

    assert isinstance(capability, PrepareOutputTools)
    assert len(contributions) == 1
    assert isinstance(contributions[0], PrepareOutputTools)
    assert [tool_def.name for tool_def in prepared_tools] == ["public"]
    assert metadata == {
        "current_mode_id": "default",
        "modes": [
            {
                "description": "Default output tools.",
                "id": "default",
                "name": "Default",
            },
            {
                "description": None,
                "id": "strict",
                "name": "Strict",
            },
        ],
    }

    mode_state = bridge.set_mode(session, Agent(TestModel()), "strict")
    assert mode_state is not None
    assert mode_state.current_mode_id == "strict"
    assert bridge.get_mode_state(session, Agent(TestModel())).current_mode_id == "strict"
    session.config_values["prepare_output_tools_mode"] = "missing"
    assert bridge.get_mode_state(session, Agent(TestModel())).current_mode_id == "default"
    assert bridge.set_mode(session, Agent(TestModel()), "missing") is None

    updates = bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None


def test_prepare_output_tools_bridge_validation_and_failure_paths(
    tmp_path: Path,
) -> None:
    with pytest.raises(ValueError, match="requires at least one mode"):
        PrepareOutputToolsBridge(default_mode_id="x", modes=[])
    with pytest.raises(ValueError, match="default mode"):
        PrepareOutputToolsBridge(
            default_mode_id="missing",
            modes=[
                PrepareOutputToolsMode(
                    id="default",
                    name="Default",
                    prepare_func=_passthrough_tools,
                ),
            ],
        )
    with pytest.raises(ValueError, match="reserved slash command names"):
        PrepareOutputToolsBridge(
            default_mode_id="model",
            modes=[
                PrepareOutputToolsMode(
                    id="model",
                    name="Model",
                    prepare_func=_passthrough_tools,
                ),
            ],
        )

    def boom(
        ctx: RunContext[None],
        tool_defs: list[ToolDefinition],
    ) -> list[ToolDefinition]:
        del ctx, tool_defs
        raise RuntimeError("output boom")

    bridge = PrepareOutputToolsBridge(
        default_mode_id="default",
        modes=[
            PrepareOutputToolsMode(
                id="default",
                name="Default",
                prepare_func=boom,
            ),
        ],
    )
    session = AcpSessionContext(
        session_id="prepare-output-tools-failure",
        cwd=tmp_path,
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )

    async def run_prepare() -> None:
        prepared = bridge.build_prepare_output_tools(session)
        result = cast("Awaitable[list[ToolDefinition]]", prepared(cast("Any", None), []))
        await result

    with pytest.raises(RuntimeError, match="output boom"):
        asyncio.run(run_prepare())
    with pytest.raises(ValueError, match="Unknown prepare output tools mode"):
        bridge._require_mode("missing")

    updates = bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None


def test_prepare_tools_bridge_allows_at_most_one_plan_mode() -> None:
    with pytest.raises(ValueError, match="at most one `plan_mode=True`"):
        PrepareToolsBridge(
            default_mode_id="chat",
            modes=[
                PrepareToolsMode(
                    id="chat",
                    name="Chat",
                    prepare_func=_passthrough_tools,
                    plan_mode=True,
                ),
                PrepareToolsMode(
                    id="plan",
                    name="Plan",
                    prepare_func=_passthrough_tools,
                    plan_mode=True,
                ),
            ],
        )


def test_prepare_tools_bridge_rejects_invalid_default_plan_generation_type() -> None:
    with pytest.raises(ValueError, match="default plan generation type"):
        PrepareToolsBridge(
            default_mode_id="plan",
            modes=[
                PrepareToolsMode(
                    id="plan",
                    name="Plan",
                    prepare_func=_passthrough_tools,
                    plan_mode=True,
                ),
            ],
            default_plan_generation_type=cast("Any", "invalid"),
        )


def test_prepare_tools_bridge_can_enable_plan_tools_outside_plan_mode() -> None:
    bridge = PrepareToolsBridge(
        default_mode_id="agent",
        modes=[
            PrepareToolsMode(
                id="plan",
                name="Plan",
                prepare_func=_passthrough_tools,
                plan_mode=True,
            ),
            PrepareToolsMode(
                id="agent",
                name="Agent",
                prepare_func=_passthrough_tools,
                plan_tools=True,
            ),
        ],
    )

    session = AcpSessionContext(
        session_id="plan-tools",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        config_values={"mode": "agent"},
    )

    assert bridge.is_plan_mode(session) is False
    assert bridge.supports_plan_tools(session) is True
    assert bridge.supports_plan_write_tools(session) is True
    assert bridge.supports_plan_progress(session) is True
    config_options = bridge.get_config_options(session, Agent(TestModel()))
    assert len(config_options) == 1
    assert config_options[0].id == "plan_generation_type"
    assert (
        bridge.set_config_option(
            session,
            Agent(TestModel()),
            "plan_generation_type",
            "tools",
        )
        is not None
    )


def test_prepare_tools_bridge_exposes_plan_generation_config_and_helpers() -> None:
    bridge = PrepareToolsBridge(
        default_mode_id="plan",
        modes=[
            PrepareToolsMode(
                id="plan",
                name="Plan",
                prepare_func=_passthrough_tools,
                plan_mode=True,
            ),
            PrepareToolsMode(
                id="agent",
                name="Agent",
                prepare_func=_passthrough_tools,
                plan_tools=True,
            ),
        ],
    )
    session = AcpSessionContext(
        session_id="plan-generation",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    agent = Agent(TestModel())

    config_options = bridge.get_config_options(session, agent)
    assert len(config_options) == 1
    assert isinstance(config_options[0], SessionConfigOptionSelect)
    assert config_options[0].id == "plan_generation_type"
    assert config_options[0].current_value == "structured"
    assert bridge.uses_structured_plan_generation(session) is True
    assert bridge.supports_plan_write_tools(session) is False
    metadata = bridge.get_session_metadata(session, agent)
    assert metadata["current_plan_generation_type"] == "structured"
    assert metadata["supported_plan_generation_types"] == ["tools", "structured"]

    updated = bridge.set_config_option(session, agent, "plan_generation_type", "tools")
    assert updated is not None
    assert bridge.current_plan_generation_type(session) == "tools"
    assert bridge.uses_tool_plan_generation(session) is True
    assert bridge.supports_plan_write_tools(session) is True
    reset = bridge.set_config_option(session, agent, "plan_generation_type", "structured")
    assert reset is not None
    assert bridge.current_plan_generation_type(session) == "structured"
    assert "plan_generation_type" not in session.config_values

    session.config_values["mode"] = "agent"
    assert bridge.supports_plan_progress(session) is True
    assert bridge.supports_plan_write_tools(session) is True
    assert bridge.set_config_option(session, agent, "plan_generation_type", True) is None
    assert bridge.set_config_option(session, agent, "plan_generation_type", "invalid") is None
    session.config_values["plan_generation_type"] = "invalid"
    assert bridge.current_plan_generation_type(session) == "structured"


def test_prepare_tools_bridge_records_failure_events(tmp_path: Path) -> None:
    def boom(
        ctx: RunContext[None],
        tool_defs: list[ToolDefinition],
    ) -> list[ToolDefinition]:
        del ctx, tool_defs
        raise RuntimeError("boom")

    bridge = PrepareToolsBridge(
        default_mode_id="plan",
        modes=[
            PrepareToolsMode(
                id="plan",
                name="Plan",
                prepare_func=boom,
                plan_mode=True,
            ),
        ],
    )
    session = AcpSessionContext(
        session_id="prepare-tools-failure",
        cwd=tmp_path,
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )

    async def run_prepare() -> None:
        prepared = bridge.build_prepare_tools(session)
        result = prepared(cast("Any", None), [])
        if asyncio.iscoroutine(result):
            await result
            return  # pragma: no cover
        assert result == []  # pragma: no cover

    with pytest.raises(RuntimeError, match="boom"):
        asyncio.run(run_prepare())

    updates = bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None


def test_prepare_tools_bridge_rejects_reserved_mode_ids() -> None:
    with pytest.raises(ValueError, match="reserved slash command names"):
        PrepareToolsBridge(
            default_mode_id="model",
            modes=[
                PrepareToolsMode(
                    id="model",
                    name="Model",
                    prepare_func=_passthrough_tools,
                ),
            ],
        )



from __future__ import annotations as _annotations

from collections.abc import Callable, Mapping, Sequence
from concurrent.futures import Executor
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Generic, Literal, TypeVar, cast
from urllib.parse import urlparse

from acp.schema import ToolKind
from pydantic_ai import ModelRequestContext
from pydantic_ai.capabilities import (
    MCP,
    AbstractCapability,
    ImageGeneration,
    IncludeToolReturnSchemas,
    PrefixTools,
    SetToolMetadata,
    ThreadExecutor,
    Toolset,
    WebFetch,
    WebSearch,
)
from pydantic_ai.messages import CompactionPart, ModelMessage, ModelResponse
from pydantic_ai.native_tools import ImageAspectRatio, WebSearchUserLocation
from pydantic_ai.tools import RunContext, ToolSelector
from pydantic_ai.toolsets import AgentToolset

from ..agent_types import RuntimeAgent
from ..projection import (
    HarnessCodeModeProjectionMap,
    HarnessFileSystemProjectionMap,
    HarnessShellProjectionMap,
    ProjectionMap,
)
from ..session.state import AcpSessionContext, JsonValue
from .base import BufferedCapabilityBridge, CapabilityBridge

if TYPE_CHECKING:
    from pydantic_ai.models import KnownModelName, Model
    from pydantic_ai.native_tools import ImageGenerationTool, MCPServerTool

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)

__all__ = (
    "AnthropicCompactionBridge",
    "HarnessCodeModeBridge",
    "HarnessFileSystemBridge",
    "HarnessShellBridge",
    "ImageGenerationBridge",
    "IncludeToolReturnSchemasBridge",
    "McpCapabilityBridge",
    "OpenAICompactionBridge",
    "PrefixToolsBridge",
    "SetToolMetadataBridge",
    "ThreadExecutorBridge",
    "ToolsetBridge",
    "WebFetchBridge",
    "WebSearchBridge",
)

_DEFAULT_WEB_SEARCH_TOOL_NAMES: Final[frozenset[str]] = frozenset(
    {
        "duckduckgo_search",
        "exa_search",
        "tavily_search",
        "web_search",
    },
)
_DEFAULT_WEB_FETCH_TOOL_NAMES: Final[frozenset[str]] = frozenset({"web_fetch"})
_DEFAULT_IMAGE_GENERATION_TOOL_NAMES: Final[frozenset[str]] = frozenset(
    {"generate_image", "image_generation"},
)
_DEFAULT_MCP_TOOL_NAME_PREFIXES: Final[frozenset[str]] = frozenset({"mcp_server:"})
_DEFAULT_HARNESS_FILESYSTEM_TOOL_NAMES: Final[frozenset[str]] = frozenset(
    {
        "create_directory",
        "edit_file",
        "list_directory",
        "read_file",
        "search_files",
        "write_file",
    },
)
_DEFAULT_HARNESS_SHELL_TOOL_NAMES: Final[frozenset[str]] = frozenset(
    {"check_command", "run_command", "start_command", "stop_command"},
)
_DEFAULT_HARNESS_CODE_MODE_TOOL_NAMES: Final[frozenset[str]] = frozenset({"run_code"})


def _missing_harness_dependency(extra: str | None = None) -> ImportError:
    suffix = f"[{extra}]" if extra else ""
    return ImportError(
        "pydantic-ai-harness is required for this bridge. "
        f"Install `pydantic-ai-harness{suffix}` alongside `pydantic-acp`.",
    )


def _json_string_list(
    values: frozenset[str] | list[str] | None,
) -> list[JsonValue] | None:
    if values is None:
        return None
    result: list[JsonValue] = []
    result.extend(sorted(values))
    return result


def _json_user_location(
    location: WebSearchUserLocation | None,
) -> dict[str, JsonValue] | None:
    if location is None:
        return None
    metadata: dict[str, JsonValue] = {}
    for key, value in location.items():
        if value is None or isinstance(value, bool | int | float | str):
            metadata[str(key)] = value
    return metadata or None


def _json_bool(value: bool) -> JsonValue:
    return value


def _json_string(value: str | None) -> JsonValue | None:
    return value.strip() if isinstance(value, str) and value.strip() else None


def _json_int(value: int | None) -> JsonValue | None:
    return value


def _resolve_mcp_server_id(url: str, explicit_id: str | None) -> str:
    if explicit_id:
        return explicit_id
    parsed = urlparse(url)
    path = parsed.path.rstrip("/")
    slug = path.split("/")[-1] if path else ""
    host = parsed.hostname or ""
    if slug:
        return f"{host}-{slug}" if host else slug
    return host or url


@dataclass(slots=True, kw_only=True)
class ThreadExecutorBridge(CapabilityBridge):
    executor: Executor
    metadata_key: str | None = "thread_executor"

    def build_capability(self, session: AcpSessionContext) -> ThreadExecutor:
        del session
        return ThreadExecutor(self.executor)

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {"executor_type": type(self.executor).__name__}


@dataclass(slots=True)
class SetToolMetadataBridge(CapabilityBridge, Generic[AgentDepsT]):
    tools: ToolSelector[AgentDepsT] = "all"
    metadata_key: str | None = None
    metadata: dict[str, JsonValue] | None = None

    def __init__(
        self,
        *,
        tools: ToolSelector[AgentDepsT] = "all",
        metadata_key: str | None = None,
        **metadata: JsonValue,
    ) -> None:
        self.tools = tools
        self.metadata_key = metadata_key
        self.metadata = dict(metadata)

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> SetToolMetadata[AgentDepsT]:
        del session
        metadata = self.metadata or {}
        return SetToolMetadata(
            tools=self.tools,
            **metadata,
        )

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)


@dataclass(slots=True)
class IncludeToolReturnSchemasBridge(CapabilityBridge, Generic[AgentDepsT]):
    tools: ToolSelector[AgentDepsT] = "all"
    metadata_key: str | None = None

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> IncludeToolReturnSchemas[AgentDepsT]:
        del session
        return IncludeToolReturnSchemas(tools=self.tools)

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)


@dataclass(slots=True, kw_only=True)
class HarnessFileSystemBridge(CapabilityBridge, Generic[AgentDepsT]):
    root_dir: str | Path = "."
    allowed_patterns: Sequence[str] = ()
    denied_patterns: Sequence[str] = ()
    protected_patterns: Sequence[str] | None = None
    max_read_lines: int = 2000
    max_search_results: int = 200
    max_find_results: int = 200
    tool_names: frozenset[str] = _DEFAULT_HARNESS_FILESYSTEM_TOOL_NAMES
    metadata_key: str | None = "harness_filesystem"

    def build_capability(self, session: AcpSessionContext) -> AbstractCapability[AgentDepsT]:
        del session
        try:
            from pydantic_ai_harness.filesystem import FileSystem
        except ImportError as exc:
            raise _missing_harness_dependency() from exc
        kwargs: dict[str, Any] = {
            "root_dir": self.root_dir,
            "allowed_patterns": list(self.allowed_patterns),
            "denied_patterns": list(self.denied_patterns),
            "max_read_lines": self.max_read_lines,
            "max_search_results": self.max_search_results,
            "max_find_results": self.max_find_results,
        }
        if self.protected_patterns is not None:
            kwargs["protected_patterns"] = list(self.protected_patterns)
        return cast("AbstractCapability[AgentDepsT]", FileSystem(**kwargs))

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_projection_maps(self) -> tuple[ProjectionMap, ...]:
        return (HarnessFileSystemProjectionMap(),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "allowed_patterns": _json_string_list(list(self.allowed_patterns)),
            "denied_patterns": _json_string_list(list(self.denied_patterns)),
            "max_find_results": self.max_find_results,
            "max_read_lines": self.max_read_lines,
            "max_search_results": self.max_search_results,
            "protected_patterns": (
                _json_string_list(list(self.protected_patterns))
                if self.protected_patterns is not None
                else None
            ),
            "root_dir": str(self.root_dir),
            "tool_names": _json_string_list(self.tool_names),
        }

    def get_tool_kind(self, tool_name: str, raw_input: JsonValue | None = None) -> ToolKind | None:
        del raw_input
        if tool_name == "read_file":
            return "read"
        if tool_name in {"write_file", "create_directory"}:
            return "edit"
        if tool_name == "edit_file":
            return "edit"
        if tool_name in {"list_directory", "search_files"}:
            return "search"
        return None


@dataclass(slots=True, kw_only=True)
class HarnessShellBridge(CapabilityBridge, Generic[AgentDepsT]):
    cwd: str | Path = "."
    allowed_commands: Sequence[str] = ()
    denied_commands: Sequence[str] | None = None
    denied_operators: Sequence[str] = ()
    default_timeout: float = 30.0
    max_output_chars: int = 50_000
    persist_cwd: bool = False
    allow_interactive: bool = False
    env: Mapping[str, str] | None = None
    denied_env_patterns: Sequence[str] = ()
    tool_names: frozenset[str] = _DEFAULT_HARNESS_SHELL_TOOL_NAMES
    metadata_key: str | None = "harness_shell"

    def build_capability(self, session: AcpSessionContext) -> AbstractCapability[AgentDepsT]:
        del session
        try:
            from pydantic_ai_harness.shell import Shell
        except ImportError as exc:
            raise _missing_harness_dependency() from exc
        kwargs: dict[str, Any] = {
            "cwd": self.cwd,
            "allowed_commands": list(self.allowed_commands),
            "denied_operators": list(self.denied_operators),
            "default_timeout": self.default_timeout,
            "max_output_chars": self.max_output_chars,
            "persist_cwd": self.persist_cwd,
            "allow_interactive": self.allow_interactive,
            "env": dict(self.env) if self.env is not None else None,
            "denied_env_patterns": list(self.denied_env_patterns),
        }
        if self.denied_commands is not None:
            kwargs["denied_commands"] = list(self.denied_commands)
        return cast("AbstractCapability[AgentDepsT]", Shell(**kwargs))

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_projection_maps(self) -> tuple[ProjectionMap, ...]:
        return (HarnessShellProjectionMap(),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "allow_interactive": self.allow_interactive,
            "allowed_commands": _json_string_list(list(self.allowed_commands)),
            "cwd": str(self.cwd),
            "default_timeout": self.default_timeout,
            "denied_commands": (
                _json_string_list(list(self.denied_commands))
                if self.denied_commands is not None
                else None
            ),
            "denied_env_patterns": _json_string_list(list(self.denied_env_patterns)),
            "denied_operators": _json_string_list(list(self.denied_operators)),
            "env_keys": _json_string_list(sorted((self.env or {}).keys())),
            "max_output_chars": self.max_output_chars,
            "persist_cwd": self.persist_cwd,
            "tool_names": _json_string_list(self.tool_names),
        }

    def get_tool_kind(self, tool_name: str, raw_input: JsonValue | None = None) -> ToolKind | None:
        del raw_input
        return "execute" if tool_name in self.tool_names else None


@dataclass(slots=True, kw_only=True)
class HarnessCodeModeBridge(CapabilityBridge, Generic[AgentDepsT]):
    tools: ToolSelector[AgentDepsT] = "all"
    max_retries: int = 3
    os_access: Any = None
    mount: Any = None
    dynamic_catalog: bool = False
    tool_names: frozenset[str] = _DEFAULT_HARNESS_CODE_MODE_TOOL_NAMES
    metadata_key: str | None = "harness_code_mode"

    def build_capability(self, session: AcpSessionContext) -> AbstractCapability[AgentDepsT]:
        del session
        try:
            from pydantic_ai_harness.code_mode import CodeMode
        except ImportError as exc:
            raise _missing_harness_dependency("code-mode") from exc
        return cast(
            "AbstractCapability[AgentDepsT]",
            CodeMode(
                tools=cast("Any", self.tools),
                max_retries=self.max_retries,
                os_access=self.os_access,
                mount=self.mount,
                dynamic_catalog=self.dynamic_catalog,
            ),
        )

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_projection_maps(self) -> tuple[ProjectionMap, ...]:
        return (HarnessCodeModeProjectionMap(),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "dynamic_catalog": self.dynamic_catalog,
            "has_mount": self.mount is not None,
            "has_os_access": self.os_access is not None,
            "max_retries": self.max_retries,
            "tool_names": _json_string_list(self.tool_names),
        }

    def get_tool_kind(self, tool_name: str, raw_input: JsonValue | None = None) -> ToolKind | None:
        del raw_input
        return "execute" if tool_name in self.tool_names else None


@dataclass(slots=True, kw_only=True)
class ImageGenerationBridge(CapabilityBridge, Generic[AgentDepsT]):
    builtin: bool | ImageGenerationTool | Any = True
    local: Any = None
    fallback_model: Model | KnownModelName | str | Callable[..., Any] | None = None
    background: Literal["transparent", "opaque", "auto"] | None = None
    input_fidelity: Literal["high", "low"] | None = None
    moderation: Literal["auto", "low"] | None = None
    output_compression: int | None = None
    output_format: Literal["png", "webp", "jpeg"] | None = None
    quality: Literal["low", "medium", "high", "auto"] | None = None
    size: Literal["auto", "1024x1024", "1024x1536", "1536x1024", "512", "1K", "2K", "4K"] | None = (
        None
    )
    aspect_ratio: ImageAspectRatio | None = None
    tool_names: frozenset[str] = _DEFAULT_IMAGE_GENERATION_TOOL_NAMES
    metadata_key: str | None = "image_generation"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> ImageGeneration[AgentDepsT]:
        del session
        return ImageGeneration(
            native=self.builtin,
            local=self.local,
            fallback_model=self.fallback_model,
            background=self.background,
            input_fidelity=self.input_fidelity,
            moderation=self.moderation,
            output_compression=self.output_compression,
            output_format=self.output_format,
            quality=self.quality,
            size=self.size,
            aspect_ratio=self.aspect_ratio,
        )

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "aspect_ratio": _json_string(self.aspect_ratio),
            "background": _json_string(self.background),
            "fallback_model": _json_string(
                self.fallback_model if isinstance(self.fallback_model, str) else None,
            ),
            "input_fidelity": _json_string(self.input_fidelity),
            "moderation": _json_string(self.moderation),
            "output_compression": _json_int(self.output_compression),
            "output_format": _json_string(self.output_format),
            "quality": _json_string(self.quality),
            "size": _json_string(self.size),
            "tool_names": _json_string_list(self.tool_names),
        }

    def get_tool_kind(
        self,
        tool_name: str,
        raw_input: JsonValue | None = None,
    ) -> Literal["execute"] | None:
        del raw_input
        return "execute" if tool_name in self.tool_names else None


@dataclass(slots=True, kw_only=True)
class McpCapabilityBridge(CapabilityBridge, Generic[AgentDepsT]):
    url: str
    builtin: bool | MCPServerTool | Any = True
    local: Any = None
    id: str | None = None
    authorization_token: str | None = None
    headers: dict[str, str] | None = None
    allowed_tools: list[str] | None = None
    description: str | None = None
    tool_name_prefixes: frozenset[str] = _DEFAULT_MCP_TOOL_NAME_PREFIXES
    metadata_key: str | None = "mcp_capability"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> MCP[AgentDepsT]:
        del session
        return MCP(
            self.url,
            native=self.builtin,
            local=self.local,
            id=self.id,
            authorization_token=self.authorization_token,
            headers=self.headers,
            allowed_tools=self.allowed_tools,
            description=self.description,
        )

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "allowed_tools": _json_string_list(self.allowed_tools),
            "description": _json_string(self.description),
            "has_authorization_token": _json_bool(self.authorization_token is not None),
            "headers": _json_string_list(sorted((self.headers or {}).keys())),
            "server_id": _resolve_mcp_server_id(self.url, self.id),
            "tool_name_prefixes": _json_string_list(self.tool_name_prefixes),
            "url": self.url,
        }

    def get_tool_kind(
        self,
        tool_name: str,
        raw_input: JsonValue | None = None,
    ) -> Literal["execute"] | None:
        del raw_input
        return (
            "execute"
            if any(tool_name.startswith(prefix) for prefix in self.tool_name_prefixes)
            else None
        )


@dataclass(slots=True, kw_only=True)
class ToolsetBridge(CapabilityBridge, Generic[AgentDepsT]):
    toolset: AgentToolset[AgentDepsT]
    metadata_key: str | None = "toolset"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> Toolset[AgentDepsT]:
        del session
        return Toolset(toolset=self.toolset)

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        toolset_id = getattr(self.toolset, "id", None)
        return {
            "toolset_id": _json_string(toolset_id if isinstance(toolset_id, str) else None),
            "toolset_type": type(self.toolset).__name__,
        }


@dataclass(slots=True, kw_only=True)
class PrefixToolsBridge(CapabilityBridge, Generic[AgentDepsT]):
    wrapped: AbstractCapability[AgentDepsT]
    prefix: str
    metadata_key: str | None = "prefix_tools"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> PrefixTools[AgentDepsT]:
        del session
        return PrefixTools(wrapped=self.wrapped, prefix=self.prefix)

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "prefix": self.prefix,
            "wrapped_capability": type(self.wrapped).__name__,
        }

    def get_tool_kind(
        self,
        tool_name: str,
        raw_input: JsonValue | None = None,
    ) -> Literal["execute"] | None:
        del raw_input
        return "execute" if tool_name.startswith(f"{self.prefix}_") else None


@dataclass(slots=True, kw_only=True)
class OpenAICompactionBridge(BufferedCapabilityBridge, Generic[AgentDepsT]):
    message_count_threshold: int | None = None
    trigger: Callable[[list[ModelMessage]], bool] | None = None
    instructions: str | None = None
    metadata_key: str | None = "openai_compaction"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> AbstractCapability[Any]:
        from pydantic_ai.models.openai import OpenAICompaction

        bridge = self

        class _BridgeOpenAICompaction(OpenAICompaction[Any]):
            def __init__(self) -> None:
                OpenAICompaction.__init__(
                    self,
                    message_count_threshold=bridge.message_count_threshold,
                    trigger=bridge.trigger,
                )

            async def before_model_request(
                self,
                ctx: RunContext[Any],
                request_context: ModelRequestContext,
            ) -> ModelRequestContext:
                if not _should_openai_compact(
                    request_context.messages,
                    trigger=self.trigger,
                    message_count_threshold=self.message_count_threshold,
                ):
                    return request_context

                event_id = bridge._record_started_event(
                    session,
                    title="Context Compaction",
                    raw_input={
                        "provider": "openai",
                        "instructions": _json_string(bridge.instructions),
                        "message_count": len(request_context.messages),
                    },
                )
                try:
                    updated_context = await OpenAICompaction.before_model_request(
                        self,
                        ctx,
                        request_context,
                    )
                except Exception as error:
                    bridge._record_progress_event(
                        session,
                        event_id=event_id,
                        title="Context Compaction",
                        status="failed",
                        raw_output=f"Provider: openai\nStatus: failed\nError: {error}",
                    )
                    raise

                bridge._record_progress_event(
                    session,
                    event_id=event_id,
                    title="Context Compaction",
                    status="completed",
                    raw_output=_format_openai_compaction_output(updated_context),
                )
                return updated_context

            @classmethod
            def get_serialization_name(cls) -> str | None:
                return "OpenAICompaction"

        return _BridgeOpenAICompaction()

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "has_trigger": _json_bool(self.trigger is not None),
            "instructions": _json_string(self.instructions),
            "message_count_threshold": _json_int(self.message_count_threshold),
        }


@dataclass(slots=True, kw_only=True)
class AnthropicCompactionBridge(CapabilityBridge, Generic[AgentDepsT]):
    token_threshold: int = 150_000
    instructions: str | None = None
    pause_after_compaction: bool = False
    metadata_key: str | None = "anthropic_compaction"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> AbstractCapability[Any]:
        del session
        from pydantic_ai.models.anthropic import AnthropicCompaction

        return AnthropicCompaction(
            token_threshold=self.token_threshold,
            instructions=self.instructions,
            pause_after_compaction=self.pause_after_compaction,
        )

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "instructions": _json_string(self.instructions),
            "pause_after_compaction": _json_bool(self.pause_after_compaction),
            "token_threshold": self.token_threshold,
        }


def _should_openai_compact(
    messages: list[ModelMessage],
    *,
    trigger: Callable[[list[ModelMessage]], bool] | None,
    message_count_threshold: int | None,
) -> bool:
    if trigger is not None:
        return trigger(messages)
    if message_count_threshold is not None:
        return len(messages) > message_count_threshold
    return False


def _format_openai_compaction_output(request_context: ModelRequestContext) -> str:
    parts: list[str] = [
        "Provider: openai",
        "Status: history compacted",
        "Compaction payload stored for round-trip.",
    ]
    compacted_part = _extract_compaction_part(request_context.messages)
    if compacted_part is not None and compacted_part.id:
        parts.append(f"Compaction id: {compacted_part.id}")
    return "\n".join(parts)


def _extract_compaction_part(messages: list[ModelMessage]) -> CompactionPart | None:
    for message in messages:
        if not isinstance(message, ModelResponse):
            continue
        for part in message.parts:
            if isinstance(part, CompactionPart):
                return part
    return None


@dataclass(slots=True, kw_only=True)
class WebSearchBridge(CapabilityBridge, Generic[AgentDepsT]):
    builtin: bool | Any = True
    local: Any = None
    search_context_size: Literal["low", "medium", "high"] | None = None
    user_location: WebSearchUserLocation | None = None
    blocked_domains: list[str] | None = None
    allowed_domains: list[str] | None = None
    max_uses: int | None = None
    tool_names: frozenset[str] = _DEFAULT_WEB_SEARCH_TOOL_NAMES
    metadata_key: str | None = "web_search"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> WebSearch[AgentDepsT]:
        del session
        return WebSearch(
            native=self.builtin,
            local=self.local,
            search_context_size=self.search_context_size,
            user_location=self.user_location,
            blocked_domains=self.blocked_domains,
            allowed_domains=self.allowed_domains,
            max_uses=self.max_uses,
        )

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "allowed_domains": _json_string_list(self.allowed_domains),
            "blocked_domains": _json_string_list(self.blocked_domains),
            "max_uses": self.max_uses,
            "search_context_size": self.search_context_size,
            "tool_names": _json_string_list(self.tool_names),
            "user_location": _json_user_location(self.user_location),
        }

    def get_tool_kind(
        self,
        tool_name: str,
        raw_input: JsonValue | None = None,
    ) -> Literal["search"] | None:
        del raw_input
        return "search" if tool_name in self.tool_names else None


@dataclass(slots=True, kw_only=True)
class WebFetchBridge(CapabilityBridge, Generic[AgentDepsT]):
    builtin: bool | Any = True
    local: Any = None
    allowed_domains: list[str] | None = None
    blocked_domains: list[str] | None = None
    max_uses: int | None = None
    enable_citations: bool | None = None
    max_content_tokens: int | None = None
    tool_names: frozenset[str] = _DEFAULT_WEB_FETCH_TOOL_NAMES
    metadata_key: str | None = "web_fetch"

    def build_capability(
        self,
        session: AcpSessionContext,
    ) -> WebFetch[AgentDepsT]:
        del session
        return WebFetch(
            native=self.builtin,
            local=self.local,
            allowed_domains=self.allowed_domains,
            blocked_domains=self.blocked_domains,
            max_uses=self.max_uses,
            enable_citations=self.enable_citations,
            max_content_tokens=self.max_content_tokens,
        )

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[AbstractCapability[Any], ...]:
        return (self.build_capability(session),)

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del session, agent
        return {
            "allowed_domains": _json_string_list(self.allowed_domains),
            "blocked_domains": _json_string_list(self.blocked_domains),
            "enable_citations": self.enable_citations,
            "max_content_tokens": self.max_content_tokens,
            "max_uses": self.max_uses,
            "tool_names": _json_string_list(self.tool_names),
        }

    def get_tool_kind(
        self,
        tool_name: str,
        raw_input: JsonValue | None = None,
    ) -> Literal["fetch"] | None:
        del raw_input
        return "fetch" if tool_name in self.tool_names else None



from __future__ import annotations as _annotations

import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, cast

import pytest
from acp import text_block
from acp.interfaces import Client
from acp.schema import (
    AgentMessageChunk,
    PermissionOption,
    RequestPermissionResponse,
    SessionNotification,
    ToolCallUpdate,
)
from acpremote import (
    CommandOptions,
    connect_remote_agent,
    serve_command,
    serve_stdio_command,
)
from acpremote import command as command_module


@dataclass(slots=True)
class _RecordingClient:
    updates: list[SessionNotification] = field(default_factory=list)

    async def request_permission(
        self,
        options: list[PermissionOption],
        session_id: str,
        tool_call: ToolCallUpdate,
        **kwargs: Any,
    ) -> RequestPermissionResponse:
        del options, session_id, tool_call, kwargs
        raise AssertionError("permission flow is not used by command relay tests")

    async def session_update(self, session_id: str, update: Any, **kwargs: Any) -> None:
        self.updates.append(
            SessionNotification(session_id=session_id, update=update, field_meta=kwargs or None),
        )

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        del method, params
        raise AssertionError("extension methods are not used by command relay tests")

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        del method, params

    def on_connect(self, conn: Any) -> None:
        del conn


@pytest.mark.asyncio
async def test_command_server_recording_client_stub_methods() -> None:
    client = _RecordingClient()

    with pytest.raises(AssertionError, match="permission flow"):
        await client.request_permission([], "session-1", cast("Any", object()))
    with pytest.raises(AssertionError, match="extension methods"):
        await client.ext_method("demo.echo", {"value": 1})

    await client.ext_notification("demo.note", {"value": 2})
    await client.session_update(
        "session-1",
        AgentMessageChunk(session_update="agent_message_chunk", content=text_block("ok")),
        source="test",
    )
    assert client.updates[0].field_meta == {"source": "test"}
    assert client.on_connect(object()) is None


def _write_stdio_acp_script(tmp_path: Path, *, emit_stderr: bool = False) -> Path:
    stderr_line = (
        '    import sys\n    print("stderr ready", file=sys.stderr)\n' if emit_stderr else ""
    )
    script_path = tmp_path / "stdio_acp_server.py"
    script_path.write_text(
        "\n".join(
            (
                "from __future__ import annotations",
                "",
                "import asyncio",
                "import os",
                "from dataclasses import dataclass",
                "from typing import Any",
                "",
                "from acp import run_agent, text_block",
                "from acp.interfaces import Client",
                "from acp.schema import AgentMessageChunk, ClientCapabilities, Implementation, InitializeResponse, NewSessionResponse, PromptResponse",
                "",
                "@dataclass(slots=True)",
                "class EchoAgent:",
                "    _conn: Client | None = None",
                "",
                "    def on_connect(self, conn: Client) -> None:",
                "        self._conn = conn",
                "",
                "    async def initialize(",
                "        self,",
                "        protocol_version: int,",
                "        client_capabilities: ClientCapabilities | None = None,",
                "        client_info: Implementation | None = None,",
                "        **kwargs: Any,",
                "    ) -> InitializeResponse:",
                "        del client_capabilities, client_info, kwargs",
                "        return InitializeResponse(protocol_version=protocol_version)",
                "",
                "    async def new_session(",
                "        self,",
                "        cwd: str,",
                "        mcp_servers: list[Any] | None = None,",
                "        **kwargs: Any,",
                "    ) -> NewSessionResponse:",
                "        del cwd, mcp_servers, kwargs",
                "        return NewSessionResponse(session_id=os.environ['ACPREMOTE_SESSION_ID'])",
                "",
                "    async def prompt(",
                "        self,",
                "        prompt: list[Any],",
                "        session_id: str,",
                "        **kwargs: Any,",
                "    ) -> PromptResponse:",
                "        del kwargs",
                "        text = ''.join(block.text for block in prompt if hasattr(block, 'text'))",
                "        payload = f\"{os.environ['ACPREMOTE_PREFIX']}:{os.getcwd()}:{text}\"",
                "        if self._conn is not None:",
                "            await self._conn.session_update(",
                "                session_id=session_id,",
                "                update=AgentMessageChunk(",
                "                    session_update='agent_message_chunk',",
                "                    content=text_block(payload),",
                "                ),",
                "            )",
                "        return PromptResponse(stop_reason='end_turn')",
                "",
                "async def main() -> None:",
                stderr_line.rstrip("\n"),
                "    await run_agent(EchoAgent())",
                "",
                "asyncio.run(main())",
            ),
        ).strip()
        + "\n",
        encoding="utf-8",
    )
    return script_path


def test_command_options_reject_empty_command() -> None:
    with pytest.raises(ValueError, match="command must not be empty"):
        CommandOptions(command=())


def test_command_env_overrides_preserve_existing_environment(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("PATH", "/usr/bin:/bin")
    monkeypatch.setenv("UNCHANGED", "keep")

    merged = command_module._build_process_env({"GOOGLE_API_KEY": "demo"})

    assert merged["PATH"] == "/usr/bin:/bin"
    assert merged["UNCHANGED"] == "keep"
    assert merged["GOOGLE_API_KEY"] == "demo"


@pytest.mark.asyncio
async def test_serve_command_relays_stdio_acp_process(tmp_path: Path) -> None:
    script_path = _write_stdio_acp_script(tmp_path)
    client = _RecordingClient()
    server = await serve_command(
        [sys.executable, str(script_path)],
        mount_path="/command",
        cwd=str(tmp_path),
        env={
            "ACPREMOTE_PREFIX": "relay",
            "ACPREMOTE_SESSION_ID": "command-session",
        },
    )
    assert server.sockets is not None
    port = next(iter(server.sockets)).getsockname()[1]
    remote = await connect_remote_agent(
        cast("Client", client),
        f"ws://127.0.0.1:{port}/command/ws",
    )
    try:
        initialized = await remote.connection.initialize(protocol_version=1)
        assert initialized.protocol_version == 1

        session = await remote.connection.new_session(cwd=str(tmp_path))
        assert session.session_id == "command-session"

        prompt_response = await remote.connection.prompt(
            [text_block("hello from ws")],
            session_id=session.session_id,
        )
        assert prompt_response.stop_reason == "end_turn"
    finally:
        await remote.close()
        server.close()
        await server.wait_closed()

    assert len(client.updates) == 1
    assert isinstance(client.updates[0].update, AgentMessageChunk)
    assert client.updates[0].update.content.text == f"relay:{tmp_path}:hello from ws"


@pytest.mark.asyncio
async def test_serve_command_metadata_exposes_remote_cwd(tmp_path: Path) -> None:
    script_path = _write_stdio_acp_script(tmp_path)
    client = _RecordingClient()
    server = await serve_command(
        [sys.executable, str(script_path)],
        mount_path="/command",
        cwd=str(tmp_path),
        env={
            "ACPREMOTE_PREFIX": "meta",
            "ACPREMOTE_SESSION_ID": "command-session",
        },
    )
    assert server.sockets is not None
    port = next(iter(server.sockets)).getsockname()[1]
    remote = await connect_remote_agent(
        cast("Client", client),
        f"ws://127.0.0.1:{port}/command/ws",
    )
    try:
        assert remote.metadata is not None
        assert remote.metadata.remote_cwd == str(tmp_path)
    finally:
        await remote.close()
        server.close()
        await server.wait_closed()


@pytest.mark.asyncio
async def test_serve_stdio_command_supports_discarded_stderr(tmp_path: Path) -> None:
    script_path = _write_stdio_acp_script(tmp_path, emit_stderr=True)
    client = _RecordingClient()
    server = await serve_stdio_command(
        CommandOptions(
            command=(sys.executable, str(script_path)),
            cwd=str(tmp_path),
            env={
                "ACPREMOTE_PREFIX": "discard",
                "ACPREMOTE_SESSION_ID": "discard-session",
            },
            stderr_mode="discard",
        ),
        mount_path="/discard",
    )
    assert server.sockets is not None
    port = next(iter(server.sockets)).getsockname()[1]
    remote = await connect_remote_agent(
        cast("Client", client),
        f"ws://127.0.0.1:{port}/discard/ws",
    )
    try:
        await remote.connection.initialize(protocol_version=1)
        session = await remote.connection.new_session(cwd=str(tmp_path))
        await remote.connection.prompt(
            [text_block("stderr hidden")],
            session_id=session.session_id,
        )
    finally:
        await remote.close()
        server.close()
        await server.wait_closed()

    assert len(client.updates) == 1
    assert isinstance(client.updates[0].update, AgentMessageChunk)
    assert client.updates[0].update.content.text == f"discard:{tmp_path}:stderr hidden"



from __future__ import annotations as _annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol, cast

from acp.schema import (
    AgentMessageChunk,
    AgentPlanUpdate,
    ContentToolCallContent,
    SessionInfoUpdate,
    TextContentBlock,
    ToolCallProgress,
    ToolCallStart,
    UserMessageChunk,
)

from .session.state import JsonValue, SessionTranscriptUpdate

__all__ = (
    "CompositeEventProjectionMap",
    "EventProjectionMap",
    "StructuredEventProjectionMap",
    "compose_event_projection_maps",
)

_EVENT_KEYS = frozenset({"acp_events", "callback_events", "events"})


class EventProjectionMap(Protocol):
    def project_event_payload(
        self,
        payload: JsonValue,
    ) -> tuple[SessionTranscriptUpdate, ...] | None: ...


@dataclass(slots=True, frozen=True, kw_only=True)
class CompositeEventProjectionMap:
    maps: tuple[EventProjectionMap, ...]

    def project_event_payload(
        self,
        payload: JsonValue,
    ) -> tuple[SessionTranscriptUpdate, ...] | None:
        updates: list[SessionTranscriptUpdate] = []
        for projection_map in self.maps:
            projected = projection_map.project_event_payload(payload)
            if projected is not None:
                updates.extend(projected)
        return tuple(updates) or None


def compose_event_projection_maps(
    projection_maps: Sequence[EventProjectionMap] | None,
) -> EventProjectionMap | None:
    if projection_maps is None:
        return None
    if len(projection_maps) == 0:
        return None
    if len(projection_maps) == 1:
        return projection_maps[0]
    return CompositeEventProjectionMap(maps=tuple(projection_maps))


@dataclass(slots=True, frozen=True, kw_only=True)
class StructuredEventProjectionMap:
    event_keys: frozenset[str] = _EVENT_KEYS

    def project_event_payload(
        self,
        payload: JsonValue,
    ) -> tuple[SessionTranscriptUpdate, ...] | None:
        event_payloads = _extract_event_payloads(payload, event_keys=self.event_keys)
        if event_payloads is None:
            return None
        projected: list[SessionTranscriptUpdate] = []
        for event_payload in event_payloads:
            projected_update = _event_payload_to_update(event_payload)
            if projected_update is not None:
                projected.append(projected_update)
        return tuple(projected) or None


def _extract_event_payloads(
    payload: JsonValue,
    *,
    event_keys: frozenset[str],
) -> tuple[dict[str, JsonValue], ...] | None:
    if isinstance(payload, list):
        projected_payloads: list[dict[str, JsonValue]] = []
        for item in payload:
            if _is_string_keyed_json_object(item):
                projected_payloads.append(cast("dict[str, JsonValue]", item))
        return tuple(projected_payloads) or None
    if not _is_string_keyed_json_object(payload):
        return None
    payload_object = cast("dict[str, JsonValue]", payload)
    for key in event_keys:
        candidate = payload_object.get(key)
        if isinstance(candidate, list):
            nested_payloads: list[dict[str, JsonValue]] = []
            for item in candidate:
                if _is_string_keyed_json_object(item):
                    nested_payloads.append(cast("dict[str, JsonValue]", item))
            return tuple(nested_payloads) or None
    if _resolve_session_update_kind(payload_object) is not None:
        return (payload_object,)
    return None


def _event_payload_to_update(
    payload: dict[str, JsonValue],
) -> SessionTranscriptUpdate | None:
    normalized = _normalize_event_payload(payload)
    session_update = _resolve_session_update_kind(normalized)
    if session_update is None:
        return None
    if session_update == "agent_message_chunk":
        content = _normalize_text_content(normalized.get("content"))
        if content is None:
            return None
        return AgentMessageChunk(
            session_update="agent_message_chunk",
            content=content,
            message_id=_optional_string(normalized.get("messageId")),
        )
    if session_update == "user_message_chunk":
        content = _normalize_text_content(normalized.get("content"))
        if content is None:
            return None
        return UserMessageChunk(
            session_update="user_message_chunk",
            content=content,
            message_id=_optional_string(normalized.get("messageId")),
        )
    if session_update == "tool_call":
        try:
            return ToolCallStart.model_validate(normalized)
        except Exception:
            return None
    if session_update == "tool_call_update":
        if isinstance(normalized.get("content"), str):
            normalized = dict(normalized)
            normalized["content"] = [
                ContentToolCallContent(
                    type="content",
                    content=TextContentBlock(type="text", text=cast("str", normalized["content"])),
                ).model_dump(mode="json", by_alias=True),
            ]
        try:
            return ToolCallProgress.model_validate(normalized)
        except Exception:
            return None
    if session_update == "session_info_update":
        try:
            return SessionInfoUpdate.model_validate(normalized)
        except Exception:
            return None
    if session_update == "plan":
        try:
            return AgentPlanUpdate.model_validate(normalized)
        except Exception:
            return None
    return None  # pragma: no cover


def _normalize_event_payload(payload: dict[str, JsonValue]) -> dict[str, JsonValue]:
    normalized = dict(payload)
    if "sessionUpdate" not in normalized:
        if "session_update" in normalized and isinstance(normalized["session_update"], str):
            normalized["sessionUpdate"] = normalized.pop("session_update")
        elif "type" in normalized and isinstance(normalized["type"], str):
            normalized["sessionUpdate"] = normalized["type"]
    return normalized


def _resolve_session_update_kind(payload: dict[str, JsonValue]) -> str | None:
    session_update = payload.get("sessionUpdate")
    if isinstance(session_update, str) and session_update in {
        "agent_message_chunk",
        "plan",
        "session_info_update",
        "tool_call",
        "tool_call_update",
        "user_message_chunk",
    }:
        return session_update
    return None


def _normalize_text_content(value: JsonValue) -> TextContentBlock | None:
    if isinstance(value, str):
        return TextContentBlock(type="text", text=value)
    if _is_string_keyed_json_object(value):
        try:
            return TextContentBlock.model_validate(value)
        except Exception:
            return None
    return None


def _optional_string(value: JsonValue) -> str | None:
    return value if isinstance(value, str) else None


def _is_string_keyed_json_object(value: JsonValue) -> bool:
    return isinstance(value, dict) and all(isinstance(key, str) for key in value)



from __future__ import annotations as _annotations

import asyncio

import pytest
from acp.exceptions import RequestError

from .support import (
    AcpSessionContext,
    AdapterConfig,
    AdapterModel,
    Agent,
    AgentMessageChunk,
    AgentPlanUpdate,
    AvailableCommandsUpdate,
    DemoConfigOptionsProvider,
    DemoModelsProvider,
    DemoModesProvider,
    DemoPlanProvider,
    JsonValue,
    MemorySessionStore,
    Path,
    RecordingClient,
    RunContext,
    SessionInfoUpdate,
    TestModel,
    ToolCallProgress,
    UserMessageChunk,
    agent_message_texts,
    create_acp_agent,
    text_block,
)


def test_custom_agent_source_is_supported(tmp_path: Path) -> None:
    class SessionAwareAgentSource:
        async def get_agent(self, session: AcpSessionContext) -> Agent[None, str]:
            return Agent(TestModel(custom_output_text=f"source:{session.cwd.name}"))

        async def get_deps(
            self,
            session: AcpSessionContext,
            agent: Agent[None, str],
        ) -> None:
            del session, agent

    adapter = create_acp_agent(
        agent_source=SessionAwareAgentSource(),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path / "source-demo"), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Use the custom agent source.")],
            session_id=session.session_id,
        ),
    )

    assert agent_message_texts(client) == ["source:source-demo"]


def test_custom_agent_source_can_supply_session_deps(tmp_path: Path) -> None:
    class SessionAwareDepsSource:
        async def get_agent(self, session: AcpSessionContext) -> Agent[int, str]:
            del session
            agent = Agent[int, str](
                TestModel(call_tools=["show_deps"], custom_output_text="deps-complete"),
                deps_type=int,
            )

            @agent.tool
            def show_deps(ctx: RunContext[int]) -> str:
                return f"deps:{ctx.deps}"

            return agent

        async def get_deps(
            self,
            session: AcpSessionContext,
            agent: Agent[int, str],
        ) -> int:
            del agent
            return len(session.cwd.name)

    adapter = create_acp_agent(
        agent_source=SessionAwareDepsSource(),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path / "deps-demo"), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Show the deps value.")],
            session_id=session.session_id,
        ),
    )

    tool_updates = [
        update.raw_output for _, update in client.updates if isinstance(update, ToolCallProgress)
    ]
    assert "deps:9" in tool_updates


def test_connected_client_is_available_to_sources_and_providers(tmp_path: Path) -> None:
    client = RecordingClient()

    class ClientAwareApprovalProvider:
        def get_approval_state(
            self,
            session: AcpSessionContext,
            agent: Agent[None, str],
        ) -> dict[str, JsonValue]:
            del agent
            return {"connected": session.client is client}

    class ClientAwareSource:
        async def get_agent(self, session: AcpSessionContext) -> Agent[None, str]:
            output_text = "client:connected" if session.client is client else "client:missing"
            return Agent(TestModel(custom_output_text=output_text))

        async def get_deps(
            self,
            session: AcpSessionContext,
            agent: Agent[None, str],
        ) -> None:
            del session, agent

    adapter = create_acp_agent(
        agent_source=ClientAwareSource(),
        config=AdapterConfig(
            approval_state_provider=ClientAwareApprovalProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path / "client-aware"), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Use the client-aware agent source.")],
            session_id=session.session_id,
        ),
    )

    session_info_updates = [
        update for _, update in client.updates if isinstance(update, SessionInfoUpdate)
    ]
    assert session_info_updates
    assert session_info_updates[-1].field_meta == {
        "pydantic_acp": {"approval_state": {"connected": True}},
    }
    assert agent_message_texts(client) == ["client:connected"]


def test_agent_factory_builds_session_specific_agents(tmp_path: Path) -> None:
    def factory(session: AcpSessionContext) -> Agent[None, str]:
        return Agent(TestModel(custom_output_text=f"factory:{session.cwd.name}"))

    adapter = create_acp_agent(
        agent_factory=factory,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    first_session = asyncio.run(adapter.new_session(cwd=str(tmp_path / "alpha"), mcp_servers=[]))
    second_session = asyncio.run(adapter.new_session(cwd=str(tmp_path / "beta"), mcp_servers=[]))

    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Use the first session agent.")],
            session_id=first_session.session_id,
        ),
    )
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Use the second session agent.")],
            session_id=second_session.session_id,
        ),
    )

    assert agent_message_texts(client) == ["factory:alpha", "factory:beta"]


def test_factory_receives_updated_session_state(tmp_path: Path) -> None:
    def factory(session: AcpSessionContext) -> Agent[None, str]:
        current_flag = str(session.config_values.get("demo_flag", "unset"))
        current_tag = str(session.metadata.get("demo_tag", "missing"))
        return Agent(TestModel(custom_output_text=f"factory:{current_flag}:{current_tag}"))

    session_store = MemorySessionStore()
    adapter = create_acp_agent(
        agent_factory=factory,
        config=AdapterConfig(session_store=session_store),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path / "stateful"), mcp_servers=[]))
    stored_session = session_store.get(session.session_id)
    assert stored_session is not None
    stored_session.config_values["demo_flag"] = "enabled"
    stored_session.metadata["demo_tag"] = "from-store"
    session_store.save(stored_session)
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Show the current session-aware factory state.")],
            session_id=session.session_id,
        ),
    )

    assert agent_message_texts(client) == ["factory:enabled:from-store"]


def test_async_agent_factory_is_supported(tmp_path: Path) -> None:
    async def factory(session: AcpSessionContext) -> Agent[None, str]:
        return Agent(TestModel(custom_output_text=f"async-factory:{session.cwd.name}"))

    adapter = create_acp_agent(
        agent_factory=factory,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path / "gamma"), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Use the async factory agent.")],
            session_id=session.session_id,
        ),
    )

    assert agent_message_texts(client) == ["async-factory:gamma"]


def test_load_missing_session_returns_none_and_resume_or_fork_raise(
    tmp_path: Path,
) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="missing-session")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )

    load_response = asyncio.run(
        adapter.load_session(cwd=str(tmp_path), session_id="missing", mcp_servers=[]),
    )
    close_response = asyncio.run(adapter.close_session(session_id="missing"))

    assert load_response is None
    assert close_response is None
    with pytest.raises(RequestError):
        asyncio.run(adapter.resume_session(cwd=str(tmp_path), session_id="missing", mcp_servers=[]))
    with pytest.raises(RequestError):
        asyncio.run(adapter.fork_session(cwd=str(tmp_path), session_id="missing", mcp_servers=[]))


def test_fork_session_clones_transcript_and_model_override(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="default", model_name="model-a")),
        config=AdapterConfig(
            allow_model_selection=True,
            available_models=[
                AdapterModel(
                    model_id="model-a",
                    name="Model A",
                    override=TestModel(custom_output_text="default", model_name="model-a"),
                ),
                AdapterModel(
                    model_id="model-b",
                    name="Model B",
                    override=TestModel(custom_output_text="switched", model_name="model-b"),
                ),
            ],
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    original = asyncio.run(adapter.new_session(cwd=str(tmp_path / "original"), mcp_servers=[]))
    asyncio.run(adapter.set_session_model(model_id="model-b", session_id=original.session_id))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Prime the original session.")],
            session_id=original.session_id,
        ),
    )

    client.updates.clear()
    forked = asyncio.run(
        adapter.fork_session(
            cwd=str(tmp_path / "forked"),
            session_id=original.session_id,
            mcp_servers=[],
        ),
    )
    resume_response = asyncio.run(
        adapter.resume_session(
            cwd=str(tmp_path / "forked"),
            session_id=forked.session_id,
            mcp_servers=[],
        ),
    )

    assert resume_response.models is not None
    assert resume_response.models.current_model_id == "model-b"
    replayed_update_types = [
        type(update)
        for _, update in client.updates
        if not isinstance(update, AvailableCommandsUpdate | SessionInfoUpdate)
    ]
    assert replayed_update_types[0] is UserMessageChunk
    assert replayed_update_types[1:] == [AgentMessageChunk] * (len(replayed_update_types) - 1)
    assert any(
        isinstance(update, SessionInfoUpdate) and update.title == "Prime the original session."
        for _, update in client.updates
    )
    assert agent_message_texts(client) == ["switched"]


def test_provider_backed_fork_preserves_session_state(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="provider:default")),
        config=AdapterConfig(
            config_options_provider=DemoConfigOptionsProvider(),
            models_provider=DemoModelsProvider(),
            modes_provider=DemoModesProvider(),
            plan_provider=DemoPlanProvider(),
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    original = asyncio.run(adapter.new_session(cwd=str(tmp_path / "original"), mcp_servers=[]))
    asyncio.run(adapter.set_session_mode(mode_id="review", session_id=original.session_id))
    asyncio.run(
        adapter.set_config_option(
            config_id="stream_enabled",
            session_id=original.session_id,
            value=True,
        ),
    )
    asyncio.run(
        adapter.set_session_model(
            model_id="provider-model-b",
            session_id=original.session_id,
        ),
    )
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Prime the provider-backed session.")],
            session_id=original.session_id,
        ),
    )

    client.updates.clear()
    forked = asyncio.run(
        adapter.fork_session(
            cwd=str(tmp_path / "forked"),
            session_id=original.session_id,
            mcp_servers=[],
        ),
    )
    resume_response = asyncio.run(
        adapter.resume_session(
            cwd=str(tmp_path / "forked"),
            session_id=forked.session_id,
            mcp_servers=[],
        ),
    )

    assert forked.models is not None
    assert forked.models.current_model_id == "provider-model-b"
    assert forked.modes is not None
    assert forked.modes.current_mode_id == "review"
    assert forked.config_options is not None
    assert [option.id for option in forked.config_options] == [
        "model",
        "mode",
        "stream_enabled",
    ]
    assert resume_response.models is not None
    assert resume_response.models.current_model_id == "provider-model-b"
    assert resume_response.modes is not None
    assert resume_response.modes.current_mode_id == "review"
    resumed_plan_updates = [
        update for _, update in client.updates if isinstance(update, AgentPlanUpdate)
    ]
    assert resumed_plan_updates
    assert [entry.content for entry in resumed_plan_updates[-1].entries] == [
        "mode:review",
        "stream:true",
    ]


def test_close_session_removes_session_from_listing(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="done")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    listed_before_close = asyncio.run(adapter.list_sessions())
    close_response = asyncio.run(adapter.close_session(session_id=session.session_id))
    listed_after_close = asyncio.run(adapter.list_sessions())
    load_after_close = asyncio.run(
        adapter.load_session(
            cwd=str(tmp_path),
            session_id=session.session_id,
            mcp_servers=[],
        ),
    )

    assert close_response is not None
    assert [item.session_id for item in listed_before_close.sessions] == [session.session_id]
    assert listed_after_close.sessions == []
    assert load_after_close is None


def test_load_session_can_skip_history_replay(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="no-replay")),
        config=AdapterConfig(
            replay_history_on_load=False,
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Prime the transcript without replay.")],
            session_id=session.session_id,
        ),
    )
    client.updates.clear()

    load_response = asyncio.run(
        adapter.load_session(cwd=str(tmp_path), session_id=session.session_id, mcp_servers=[]),
    )

    assert load_response is not None
    assert all(
        isinstance(update, AvailableCommandsUpdate | SessionInfoUpdate)
        for _, update in client.updates
    )
    assert any(
        isinstance(update, SessionInfoUpdate)
        and update.title == "Prime the transcript without replay."
        for _, update in client.updates
    )



from __future__ import annotations as _annotations

from ._projection_risk import caution_for_command, caution_for_path
from ._projection_text import (
    DEFAULT_TEXT_TRUNCATION_MARKER,
    format_code_block,
    format_diff_preview,
    format_terminal_status,
    single_line_summary,
    truncate_lines,
    truncate_text,
)

__all__ = (
    "DEFAULT_TEXT_TRUNCATION_MARKER",
    "caution_for_command",
    "caution_for_path",
    "format_code_block",
    "format_diff_preview",
    "format_terminal_status",
    "single_line_summary",
    "truncate_lines",
    "truncate_text",
)
